﻿// <copyright file="DoubleBufferedTextReader.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the DoubleBufferedTextReader Class</summary>

namespace Core.IO
{
    using System;
    using System.Collections.Generic;
    using System.Text;
    using System.IO;

    /// <summary>
    /// A TextReader object that implements a Double buffer around another TextReader object.
    /// </summary>
    public sealed class DoubleBufferedTextReader : TextReader
    {
        private TextReader innerTextReader;
        private int bufferSize;

        public DoubleBufferedTextReader(TextReader innerTextReader, int bufferSize)
        {

            if (innerTextReader == null)
            {
                throw new ArgumentNullException("innerTextReader");
            }

            if (bufferSize <= 0)
            {
                throw new ArgumentOutOfRangeException("bufferSize", bufferSize, "The bufferSize must be a positive number.");
            }
            
            this.innerTextReader = innerTextReader;
            this.bufferSize = bufferSize;
        }

        /// <summary>
        /// Gets the size of the buffer.
        /// </summary>
        public int BufferSize
        {
            get
            {
                return this.bufferSize;
            }
        }

        public override void Close()
        {
            this.innerTextReader.Close();
        }

        public override int Peek()
        {
            return this.innerTextReader.Peek();
        }

        public override int Read()
        {
            return this.innerTextReader.Read();
        }
        
        public override int Read(char[] buffer, int index, int count)
        {
            return this.innerTextReader.Read(buffer, index, count);
        }
        
        public override int ReadBlock(char[] buffer, int index, int count)
        {
            return this.innerTextReader.ReadBlock(buffer, index, count);
        }
        
        public override string ReadLine()
        {
            return this.innerTextReader.ReadLine();
        }

        public override string ReadToEnd()
        {
            return this.innerTextReader.ReadToEnd();
        }
    }
}
