﻿// <copyright file="JsonReader.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the JsonReader Class</summary>

namespace Core.Json
{
    using System.IO;
    using System.Text;
    using System.Globalization;
    using System;

    public sealed class JsonReader
    {
        private StringBuilder builder = new StringBuilder();
        private TextReader innerTextReader;
        private long location;
        private bool eoi;

        public JsonReader()
        {
        }

        public TextReader InnerTextReader
        {
            get
            {
                return this.innerTextReader;
            }
            set
            {
                this.innerTextReader = value;
            }
        }

        public bool EOI
        {
            get
            {
                return eoi;
            }
        }

        private int Read()
        {
            if (!this.eoi)
            {
                int result = innerTextReader.Read();
                
                if (result == -1)
                {
                    this.eoi = true;
                }

                this.location += 1;

                return result;
            }
            else
            {
                return -1;
            }
        }

        private void Unread(int c)
        {

        }

        public void FillToken(JsonToken token)
        {
            int c = this.Read();
            while (c != -1)
            {
                long l = this.location;
                switch (c)
                {
                    case 0x09:
                    case 0xA0:
                    case 0xD0:
                    case 0x20:
                        c = this.Read();
                        break;
                    case 0x22: /* " */
                        c = this.Read();
                        while (c != -1)
                        {
                            if (c == 0x5C) /* \ */
                            {
                                c = this.Read();
                                switch (c)
                                {
                                    case 0x22: /* " */
                                        builder.Append('"');
                                        c = this.Read();
                                        break;
                                    case 0x2F: /* / */
                                        builder.Append('/');
                                        c = this.Read();
                                        break;
                                    case 0x5C: /* \ */
                                        builder.Append('\\');
                                        c = this.Read();
                                        break;
                                    case 0x62: /* b */
                                        builder.Append('\b');
                                        c = this.Read();
                                        break;
                                    case 0x66: /* f */
                                        builder.Append('\f');
                                        c = this.Read();
                                        break;
                                    case 0x6E: /* n */
                                        builder.Append('\n');
                                        c = this.Read();
                                        break;
                                    case 0x72: /* t */
                                        builder.Append('\r');
                                        c = this.Read();
                                        break;
                                    case 0x74: /* t */
                                        builder.Append('\t');
                                        c = this.Read();
                                        break;
                                    case 0x75: /* u */
                                        int d0 = this.Read();
                                        if ((d0 >= 0x30 && d0 <= 0x39) || (d0 >= 0x41 && d0 <= 0x46)|| (d0 >= 0x61 && d0 <= 0x66))
                                        {
                                            int d1 = this.Read();
                                            if ((d1 >= 0x30 && d1 <= 0x39)|| (d1 >= 0x41 && d1 <= 0x46)|| (d1 >= 0x61 && d1 <= 0x66))
                                            {
                                                int d2 = this.Read();
                                                if ((d2 >= 0x30 && d2 <= 0x39)|| (d2 >= 0x41 && d2 <= 0x46)|| (d2 >= 0x61 && d2 <= 0x66))
                                                {
                                                    int d3 = this.Read();
                                                    if ((d3 >= 0x30 && d3 <= 0x39) || (d3 >= 0x41 && d3 <= 0x46)|| (d3 >= 0x61 && d3 <= 0x66))
                                                    {
                                                        d0 = ((d0 >= 0x30 && d0 <= 0x39) ? d0 - 0x30 : (d0 >= 0x41 && d0 <= 0x46) ? d0 - 0x41 + 0x0A : d0 - 0x61 + 0x0A) << 12;
                                                        d1 = ((d1 >= 0x30 && d1 <= 0x39) ? d1 - 0x30 : (d1 >= 0x41 && d1 <= 0x46) ? d1 - 0x41 + 0x0A : d1 - 0x61 + 0x0A) << 8;
                                                        d2 = ((d2 >= 0x30 && d2 <= 0x39) ? d2 - 0x30 : (d2 >= 0x41 && d2 <= 0x46) ? d2 - 0x41 + 0x0A : d2 - 0x61 + 0x0A) << 4;
                                                        d3 = (d3 >= 0x30 && d3 <= 0x39) ? d3 - 0x30 : (d3 >= 0x41 && d3 <= 0x46) ? d3 - 0x41 + 0x0A : d3 - 0x61 + 0x0A;
                                                        builder.Append((char)(d0 + d1 + d2 + d3));
                                                        c = this.Read();
                                                        break;
                                                    }
                                                    else
                                                    {
                                                        token.Location = l;
                                                        token.Lexeme = builder.ToString();
                                                        token.TokenId = JsonTokenId.Error;
                                                        return;
                                                    }
                                                }
                                                else
                                                {
                                                    token.Location = l;
                                                    token.Lexeme = builder.ToString();
                                                    token.TokenId = JsonTokenId.Error;
                                                    return;
                                                }
                                            }
                                            else
                                            {
                                                token.Location = l;
                                                token.Lexeme = builder.ToString();
                                                token.TokenId = JsonTokenId.Error;
                                                return;
                                            }
                                        }
                                        else
                                        {
                                            token.Location = l;
                                            token.Lexeme = builder.ToString();
                                            token.TokenId = JsonTokenId.Error;
                                            return;
                                        }
                                    default:
                                        token.Location = l;
                                        token.Lexeme = builder.ToString();
                                        token.TokenId = JsonTokenId.Error;
                                        return;
                                }
                            }
                            else if (c == 0x22)
                            {
                                token.Location = l;
                                token.Lexeme = builder.ToString();
                                token.TokenId = JsonTokenId.StringLiteral;
                            }
                            else
                            {
                                builder.Append((char)c);
                                c = this.Read();
                            }
                        }
                        break;
                    case 0x2C: /* , */
                        token.Location = l;
                        token.Lexeme = null;
                        token.TokenId = JsonTokenId.Comma;
                        return;
                    case 0x3A: /* : */
                        token.Location = l;
                        token.Lexeme = null;
                        token.TokenId = JsonTokenId.Colon;
                        return;
                    case 0x5B: /* [ */
                        token.Location = l;
                        token.Lexeme = null;
                        token.TokenId = JsonTokenId.OpeningBrackets;
                        return;
                    case 0x5D: /* ] */
                        token.Location = l;
                        token.Lexeme = null;
                        token.TokenId = JsonTokenId.ClosingBrackets;
                        return;
                    case 0x66: /* f */
                        c = this.Read();
                        if (c == 0x61) /* a */
                        {
                            c = this.Read();
                            if (c == 0x6C) /* l */
                            {
                                c = this.Read();
                                if (c == 0x73) /* s */
                                {
                                    c = this.Read();
                                    if (c == 0x65) /* e */
                                    {
                                        this.builder.Append((char)c);
                                        token.Location = l;
                                        token.Lexeme = this.builder.ToString();
                                        token.TokenId = JsonTokenId.FalseLiteral;
                                        return;
                                    }
                                }
                            }
                        }

                        token.Location = l;
                        token.Lexeme = char.ConvertFromUtf32(c);
                        token.TokenId = JsonTokenId.Error;
                        return;
                    case 0x6E: /* n */
                        this.builder.Append((char)c);
                        c = this.Read();
                        if (c == 0x75) /* u */
                        {
                            this.builder.Append((char)c);
                            c = this.Read();
                            if (c == 0x6C) /* l */
                            {
                                this.builder.Append((char)c);
                                c = this.Read();
                                if (c == 0x6C) /* l */
                                {
                                    this.builder.Append((char)c);
                                    token.Location = l;
                                    token.Lexeme = this.builder.ToString();
                                    token.TokenId = JsonTokenId.NullLiteral;
                                    return;
                                }
                            }
                        }

                        this.Unread((char)c);
                        token.Location = l;
                        token.Lexeme = this.builder.ToString();
                        token.TokenId = JsonTokenId.Error;
                        return;
                    case 0x74: /* t */
                        this.builder.Append((char)c);
                        c = this.Read();
                        if (c == 0x72) /* r */
                        {
                            this.builder.Append((char)c);
                            c = this.Read();
                            if (c == 0x75) /* u */
                            {
                                this.builder.Append((char)c);
                                c = this.Read();
                                if (c == 0x65) /* e */
                                {
                                    this.builder.Append((char)c);
                                    token.Location = l;
                                    token.Lexeme = this.builder.ToString();
                                    token.TokenId = JsonTokenId.TrueLiteral;
                                    return;
                                }
                            }
                        }

                        this.Unread(c);
                        token.Location = l;
                        token.Lexeme = this.builder.ToString();
                        token.TokenId = JsonTokenId.Error;
                        return;
                    case 0x7B:
                        token.Location = l;
                        token.Lexeme = null;
                        token.TokenId = JsonTokenId.OpeningBraces;
                        return;
                    case 0x7D:
                        token.Location = l;
                        token.Lexeme = null;
                        token.TokenId = JsonTokenId.ClosingBraces;
                        return;
                    default:
                        if (c == 0x2D)
                        {
                            this.builder.Append((char)c);
                            c = this.Read();
                        }

                        if (c == 0x30)
                        {
                            this.builder.Append((char)c);
                            c = this.Read();
                        }
                        else
                        {
                            if (c >= 0x31 && c <= 0x39)
                            {
                                this.builder.Append((char)c);
                                c = this.Read();

                                while (c >= 0x30 && c <= 0x39) /* 0 - 9 */
                                {
                                    this.builder.Append((char)c);
                                    c = this.Read();
                                }
                            }
                            else
                            {
                                if (c == -1)
                                {
                                    token.Location = l;
                                    token.Lexeme = null;
                                    token.TokenId = JsonTokenId.EndOfInput;
                                    return;
                                }
                                else
                                {
                                    this.Unread(c);
                                    token.Location = l;
                                    token.Lexeme = builder.ToString();
                                    token.TokenId = JsonTokenId.Error;
                                    return;
                                }
                            }
                        }

                        if (c == 0x2E) /* . */
                        {
                            this.builder.Append((char)c);
                            c = this.Read();
                            if (c >= 0x30 && c <= 0x39) /* 0-9 */
                            {
                                this.builder.Append((char)c);
                                c = this.Read();
                                while (c >= 0x30 && c <= 0x39) /* 0-9 */
                                {
                                    this.builder.Append((char)c);
                                    c = this.Read();
                                }

                                if (c == 0x45 || c == 0x65) /* E or e */
                                {
                                    this.builder.Append((char)c);
                                    c = this.Read();

                                    if (c == 0x2B || c == 0x2D) /* + or - */
                                    {
                                        this.builder.Append((char)c);
                                        c = this.Read();
                                    }

                                    if (c >= 0x30 && c <= 0x39) /* 0-9 */
                                    {
                                        this.builder.Append((char)c);
                                        c = this.Read();
                                        while (c >= 0x30 && c <= 0x39) /* 0-9 */
                                        {
                                            this.builder.Append((char)c);
                                            c = this.Read();
                                        }

                                        this.Unread(c);
                                        token.Location = l;
                                        token.Lexeme = this.builder.ToString();
                                        token.TokenId = JsonTokenId.NumberLiteral;
                                        return;
                                    }
                                    else
                                    {
                                        this.Unread(c);
                                        token.Location = l;
                                        token.Lexeme = this.builder.ToString();
                                        token.TokenId = JsonTokenId.Error;
                                        return;
                                    }

                                }
                                else
                                {
                                    this.Unread(c);
                                    token.Location = l;
                                    token.Lexeme = this.builder.ToString();
                                    token.TokenId = JsonTokenId.NumberLiteral;
                                    return;
                                }
                            }
                            else
                            {
                                this.Unread(c);
                                token.Location = l;
                                token.Lexeme = this.builder.ToString();
                                token.TokenId = JsonTokenId.Error;
                                return;
                            }
                        }
                        else
                        {
                            this.Unread(c);
                            token.Location = l;
                            token.Lexeme = this.builder.ToString();
                            token.TokenId = JsonTokenId.NumberLiteral;
                            return;
                        }

                }
            }
        }
    }
}
