﻿// <copyright file="JsonWriter.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the JsonWriter Class</summary>

namespace Core.Json
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Globalization;
    using System.IO;
    using Core.Json.Serializers;
    using Core.Reflection.TypeExpressions;

    public class JsonWriter
    {
        private TextWriter innerTextWriter;

        public TextWriter InnerTextWriter 
        {
            get
            {
                return this.innerTextWriter;
            }
            set
            {
                if (value == null)
                {
                    innerTextWriter = TextWriter.Null;
                }
                else
                {
                    innerTextWriter = value;
                }
            }
        }

        private List<ITypeSerializer> typeSerializers = new List<ITypeSerializer>();
        
        public IList<ITypeSerializer> TypeSerializers
        {
            get
            {
                return this.typeSerializers;
            }
        }

        /// <summary>
        /// Initializes a new instance of the JsonWriter class.
        /// </summary>
        public JsonWriter()
        {
            this.innerTextWriter = TextWriter.Null;
            this.typeSerializers.Add(CommonSerializers.BooleanSerializer);
            this.typeSerializers.Add(CommonSerializers.CharSerializer);
            this.typeSerializers.Add(CommonSerializers.DateTimeSerializer);
            this.typeSerializers.Add(CommonSerializers.TimeSpanSerializer);
            this.typeSerializers.Add(CommonSerializers.GuidSerializer);
            this.typeSerializers.Add(CommonSerializers.IntegerTypeSerializer);
            this.typeSerializers.Add(CommonSerializers.StringSerializer);
            this.typeSerializers.Add(CommonSerializers.SingleSerializer);
            this.typeSerializers.Add(CommonSerializers.DoubleSerializer);
            this.typeSerializers.Add(CommonSerializers.DecimalSerializer);
            this.typeSerializers.Add(CommonSerializers.DataRowSerializer);
            this.typeSerializers.Add(CommonSerializers.DataTableSerializer);
            this.typeSerializers.Add(CommonSerializers.DataSetSerializer);
            this.typeSerializers.Add(CommonSerializers.DBNullSerializer);
            this.typeSerializers.Add(CommonSerializers.IEnumerableSerializer);
            this.typeSerializers.Add(CommonSerializers.ObjectSerializer);
        }

        public void Serialize(object source)
        {
            if (source == null)
            {
                this.WriteVerbatim("null");
            }
            else
            {
                TypeExpressionEvaluationVisitor visitor = new TypeExpressionEvaluationVisitor();
                Dictionary<string, Type> dic = new Dictionary<string, Type>();
                dic.Add("type", source.GetType());

                foreach (ITypeSerializer serializer in this.typeSerializers)
                {
                    if (visitor.Evaluate(serializer.TypeExpression, dic))
                    {
                        serializer.Serialize(source, this);
                        return;
                    }
                }

                throw new InvalidOperationException("A compatible ITypeSerializer instance is not present for the source object.");
            }
        }

        public virtual void WriteArray(IEnumerable enumerable)
        {

            if (enumerable == null)
            {
                throw new ArgumentNullException("enumerable");
            }

            IEnumerator enumerator = enumerable.GetEnumerator();

            this.WriteOpeningBrackets();

            enumerator.Reset();

            if (enumerator.MoveNext())
            {
                this.Serialize(enumerator.Current);
                while (enumerator.MoveNext())
                {
                    this.WriteComma();
                    this.Serialize(enumerator.Current);
                }
            }

            this.WriteClosingBrackets();
        }

        public virtual void WriteProperty(string propertyName, object source)
        {
            if (propertyName == null)
            {
                throw new ArgumentNullException("propertyName");
            }

            this.WriteVerbatimString(propertyName);
            this.WriteColon();
            this.Serialize(source);
        }

        public virtual void WriteOpeningBraces()
        {
            this.InnerTextWriter.Write('{');
        }

        public virtual void WriteOpeningBrackets()
        {
            this.InnerTextWriter.Write('[');
        }

        public virtual void WriteClosingBraces()
        {
            this.InnerTextWriter.Write('}');
        }

        public virtual void WriteClosingBrackets()
        {
            this.InnerTextWriter.Write(']');
        }

        public virtual void WriteComma()
        {
            this.InnerTextWriter.Write(',');
        }

        public virtual void WriteColon()
        {
            this.InnerTextWriter.Write(':');
        }

        public virtual void WriteDoubleQuotationMarks()
        {
            this.InnerTextWriter.Write('\"');
        }

        public virtual void WriteVerbatimString(string value)
        {
            this.InnerTextWriter.Write('\"');
            this.InnerTextWriter.Write(value);
            this.InnerTextWriter.Write('\"');
        }

        public virtual void WriteVerbatimString(char value)
        {
            this.InnerTextWriter.Write('\"');
            this.InnerTextWriter.Write(value);
            this.InnerTextWriter.Write('\"');
        }

        public virtual void WriteEscapedString(string value)
        {
            this.InnerTextWriter.Write('\"');
            this.WriteEscaped(value);
            this.InnerTextWriter.Write('\"');
        }

        public virtual void WriteEscapedString(char[] value)
        {
            this.InnerTextWriter.Write('\"');
            this.WriteEscaped(value);
            this.InnerTextWriter.Write('\"');
        }

        public virtual void WriteEscapedString(char value)
        {
            this.InnerTextWriter.Write('\"');
            this.WriteEscaped(value);
            this.InnerTextWriter.Write('\"');
        }

        public virtual void WriteVerbatim(string value)
        {
            this.InnerTextWriter.Write(value);
        }

        public virtual void WriteVerbatim(char value)
        {
            this.InnerTextWriter.Write(value);
        }

        public virtual void WriteEscaped(string value)
        {
            for (int i = 0; i < value.Length; i++)
            {
                this.WriteEscaped(value[i]);
            }
        }

        public virtual void WriteEscaped(char[] value)
        {
            for (int i = 0; i < value.Length; i++)
            {
                this.WriteEscaped(value[i]);
            }
        }

        public virtual void WriteEscaped(char value)
        {
            switch (value)
            {
                case '\x00':
                case '\x01':
                case '\x02':
                case '\x03':
                case '\x04':
                case '\x05':
                case '\x06':
                case '\x07':
                case '\x0B':
                case '\x0E':
                case '\x0F':
                case '\x10':
                case '\x11':
                case '\x12':
                case '\x13':
                case '\x14':
                case '\x15':
                case '\x16':
                case '\x17':
                case '\x18':
                case '\x19':
                case '\x1A':
                case '\x1B':
                case '\x1C':
                case '\x1D':
                case '\x1E':
                case '\x1F':
                    this.WriteVerbatim(string.Format(CultureInfo.InvariantCulture, "\\u{0:X4}", Convert.ToInt32(value)));
                    break;
                case '\b': // '\x08'
                    this.WriteVerbatim("\\b");
                    break;
                case '\t': // '\x09'
                    this.WriteVerbatim("\\t");
                    break;
                case '\n': // '\x0A'
                    this.WriteVerbatim("\\n");
                    break;
                case '\f': // '\x0C'
                    this.WriteVerbatim("\\f");
                    break;
                case '\r': // '\x0D'
                    this.WriteVerbatim("\\r");
                    break;
                case '\"': // '\x22'
                    this.WriteVerbatim("\\\"");
                    break;
                case '/': // '\x2F'
                    this.WriteVerbatim("\\/");
                    break;
                case '\\': // '\x5C'
                    this.WriteVerbatim("\\\\");
                    break;
                default:
                    this.WriteVerbatim(value);
                    break;
            }
        }
    }
}
