﻿// <copyright file="SqlServerTransactionLog.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the SqlServerTransactionLog Class</summary>

namespace Log
{
    using System;
    using System.Collections.Generic;
    using System.Data;
    using System.Data.SqlClient;
    using System.Globalization;
    using Core.Database;

    public sealed class SqlServerTransactionLog : ITransactionLog
    {
        public SqlConnection Connection { get; private set; }
        public string TransactionTableName { get; private set; }

        private SqlServerDatabaseHelper dbHelper = new SqlServerDatabaseHelper();

        public SqlServerTransactionLog(SqlConnection connection, string transactionTableName)
        {

            if (connection == null)
            {
                throw new ArgumentNullException("connection");
            }

            if (transactionTableName == null)
            {
                throw new ArgumentNullException("transactionTableName");
            }
                
            this.Connection = connection;
            this.TransactionTableName = transactionTableName;
        }

        public ITransactionData LogTransactionStart(string transactionType, string transactionStartData)
        {
            this.CheckConnection();

            SqlServerTransactionData result = new SqlServerTransactionData { TransactionType = transactionType, TransactionStartData = transactionStartData, TransactionStartDateTime = DateTime.Now };

            using (SqlTransaction t = this.Connection.BeginTransaction(IsolationLevel.Serializable))
            {
                result.TransactionId = (long)dbHelper.ExecuteScalarCommandText(t, string.Format(CultureInfo.InvariantCulture, "SELECT COUNT(*) + 1 FROM {0}", this.TransactionTableName));

                SqlParameter paramTransactionId = SqlServerDatabaseHelper.CreateBigIntParameter("TransactionId", result.TransactionId);
                SqlParameter paramTransactionType = SqlServerDatabaseHelper.CreateNVarCharParameter("TransactionType", result.TransactionType, 256);
                SqlParameter paramTransactionStartData = SqlServerDatabaseHelper.CreateNTextParameter("TransactionStartData", result.TransactionStartData);
                SqlParameter paramTransactionStartDateTime = SqlServerDatabaseHelper.CreateDateTimeParameter("TransactionStartDateTime", result.TransactionStartDateTime);
                
                dbHelper.ExecuteNonQueryCommandText(
                    t, 
                    string.Format(
                        CultureInfo.InvariantCulture, 
                        "INSERT INTO {0} (TransactionId, TransactionType, TransactionStartDateTime, TransactionStartData) VALUES (@TransactionId, @TransactionType, @TransactionStartDateTime, @TransactionStartData)", 
                        this.TransactionTableName),
                    paramTransactionId,
                    paramTransactionType,
                    paramTransactionStartData,
                    paramTransactionStartDateTime);
                
                t.Commit();
            }

            return result;
        }

        public ITransactionData LogTransactionEnd(ITransactionData transaction, string transactionEndData)
        {
            this.CheckConnection();


            if (transaction == null)
            {
                throw new ArgumentNullException("transaction");
            }
            else
            {
                SqlServerTransactionData tran = transaction as SqlServerTransactionData;
                tran.TransactionEndData = transactionEndData;
                tran.TransactionEndDateTime = DateTime.Now;
            }

            using (SqlTransaction t = this.Connection.BeginTransaction(IsolationLevel.Serializable))
            {
                SqlParameter paramTransactionId = SqlServerDatabaseHelper.CreateDateTimeParameter("TransactionId", transaction.TransactionId);
            
                bool exists = ((int)dbHelper.ExecuteScalarCommandText(t, string.Format(CultureInfo.InvariantCulture, "SELECT COUNT(*) FROM {0} TransactionId = @TransactionId", this.TransactionTableName), paramTransactionId)) == 1;
                
                if (exists)
                {
                    bool finished = ((int)dbHelper.ExecuteScalarCommandText(t, string.Format(CultureInfo.InvariantCulture, "SELECT COUNT(*) FROM {0} TransactionId = @TransactionId AND TransactionEndDateTime IS NULL", this.TransactionTableName), paramTransactionId)) == 0;
                    
                    if (!finished)
                    {
                        SqlParameter paramTransactionEndDateTime = SqlServerDatabaseHelper.CreateDateTimeParameter("TransactionEndDataTime", transaction.TransactionEndDateTime);
                        SqlParameter paramTransactionEndData = SqlServerDatabaseHelper.CreateNTextParameter("TransactionEndData", transaction.TransactionEndData);
                        dbHelper.ExecuteNonQueryCommandText(
                            t, 
                            string.Format(
                                CultureInfo.InvariantCulture,
                                "UPDATE {0} SET TransactionEndDateTime = @TransactionEndDateTime, TransactionEndData = @TransactionEndData WHERE TransactionId = @TransactionId", 
                                this.TransactionTableName), 
                            paramTransactionId, 
                            paramTransactionEndDateTime, 
                            paramTransactionEndData);
                    }
                    else
                    {
                        t.Rollback();
                        throw new InvalidOperationException("Transaction already finished");
                    }
                }
                else
                {
                    t.Rollback();
                    throw new InvalidOperationException("Transaction does not exists");
                }

                t.Commit();
            }

            return transaction;
        }

        public long GetTransactionsCount()
        {
            this.CheckConnection();

            return (long)dbHelper.ExecuteScalarCommandText(
                this.Connection,
                string.Format(
                    CultureInfo.InvariantCulture, 
                    "SELECT COUNT(*) FROM {0}", 
                    this.TransactionTableName));
        }

        public ITransactionData GetTransactionByTransactionId(long transactionId)
        {
            this.CheckConnection();

            SqlParameter paramTransactionId = SqlServerDatabaseHelper.CreateBigIntParameter("TransactionId", transactionId);
            
            IDataReader reader = dbHelper.ExecuteReaderCommandText(
                this.Connection,
                string.Format(
                    CultureInfo.InvariantCulture,
                    "SELECT * FROM {0} WHERE TransactionId = @TransactionId",
                    this.TransactionTableName),
                paramTransactionId);

            return MaterializeTransactionData(reader as IDataRecord);
        }

        public IEnumerable<ITransactionData> GetTransactions()
        {
            this.CheckConnection();

            IDataReader reader = dbHelper.ExecuteReaderCommandText(
                this.Connection,
                string.Format(
                    CultureInfo.InvariantCulture,
                    "SELECT * FROM {0} ORDER BY TransactionId",
                    this.TransactionTableName));

            return MaterializeTransactionData(reader);
        }

        public IEnumerable<ITransactionData> GetTransactionsByTransactionType(string transactionType)
        {
            this.CheckConnection();

            SqlParameter paramTransactionType = SqlServerDatabaseHelper.CreateNVarCharParameter("TransactionType", transactionType, 256);

            IDataReader reader = dbHelper.ExecuteReaderCommandText(
                this.Connection,
                string.Format(
                    CultureInfo.InvariantCulture,
                    "SELECT * FROM {0} WHERE TransactionType = @TransactionType ORDER BY TransactionId",
                    this.TransactionTableName),
                paramTransactionType);

            return MaterializeTransactionData(reader);
        }

        public IEnumerable<ITransactionData> GetTransactionsNotEnded()
        {
            this.CheckConnection();

            IDataReader reader = dbHelper.ExecuteReaderCommandText(
                this.Connection,
                string.Format(
                    CultureInfo.InvariantCulture,
                    "SELECT * FROM {0} WHERE TransactionEndDateTime IS NULL ORDER BY TransactionId",
                    this.TransactionTableName));

            return MaterializeTransactionData(reader);
        }

        public void Dispose()
        {
        }

        private ITransactionData MaterializeTransactionData(IDataRecord record)
        {
            SqlServerTransactionData result = new SqlServerTransactionData();

            result.TransactionId = Convert.ToInt64(record["TransactionID"]);
            
            if (record.IsDBNull(record.GetOrdinal("TransactionType")))
            {
                result.TransactionType = null;
            }
            else
            {
                result.TransactionType = Convert.ToString(record["TransactionType"]);
            }

            result.TransactionStartDateTime = Convert.ToDateTime(record["TransactionStartDateTime"]);

            if (record.IsDBNull(record.GetOrdinal("TransactionEndDateTime")))
            {
                result.TransactionEndDateTime = null;
            }
            else
            {
                result.TransactionEndDateTime = Convert.ToDateTime(record["TransactionEndDateTime"]);
            }

            if (record.IsDBNull(record.GetOrdinal("TransactionStartData")))
            {
                result.TransactionStartData = null;
            }
            else
            {
                result.TransactionStartData = Convert.ToString(record["TransactionStartData"]);
            }

            if (record.IsDBNull(record.GetOrdinal("TransactionEndData")))
            {
                result.TransactionEndData = null;
            }
            else
            {
                result.TransactionEndData = Convert.ToString(record["TransactionEndData"]);
            }
            
            return result;
        }

        private List<ITransactionData> MaterializeTransactionData(IDataReader reader)
        {
            List<ITransactionData> list = new List<ITransactionData>();
            IDataRecord record = reader as IDataRecord;
            while (reader.Read())
            {
                list.Add(MaterializeTransactionData(record));
            }
            return list;
        }

        private void CheckConnection()
        {
            if (Connection.State == System.Data.ConnectionState.Closed)
            {
                throw new InvalidOperationException();
            }
        }
    }
}
