﻿// <copyright file="ObjectReadablePropertyCursor.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ObjectReadablePropertyCursor Class</summary>

namespace Core.Reflection.Readers
{
    using System;
    using System.Reflection;

    public sealed class ObjectReadablePropertyCursor : IReadablePropertyCursor
    {
        private PropertyInfo[] properties;
        private FieldInfo[] fields;
        private int propertiesIndex = -1;
        private int fieldsIndex = -1;

        internal ObjectReadablePropertyCursor(object obj)
	    {
            Type t = obj.GetType();
            this.Object = obj;
            this.properties = t.GetProperties(BindingFlags.Public | BindingFlags.Instance);
            this.fields = t.GetFields(BindingFlags.Public | BindingFlags.Instance);
	    }

        public Object Object
        {
            get;
            private set;
        }

        public bool MoveToNextProperty()
        {
            if (this.propertiesIndex < this.properties.Length)
            {
                do
                {
                    this.propertiesIndex++;
                }
                while (this.propertiesIndex < this.properties.Length && !this.properties[this.propertiesIndex].CanRead);

                if (this.propertiesIndex < this.properties.Length)
                {
                    return true;
                }
            }
            
            if (this.fieldsIndex < this.fields.Length)
            {
                this.fieldsIndex++;
                if (this.fieldsIndex < this.fields.Length)
                {
                    return true;
                }
            }

            return false;
        }

        private static InvalidOperationException NewCursorDepletedException()
        {
            return new InvalidOperationException("The cursor has read the last object.");
        }

        public Type CurrentPropertyType
        {
            get
            {
                if (propertiesIndex == -1)
                {
                    throw new InvalidOperationException();
                }
                if (this.propertiesIndex < this.properties.Length)
                {
                    return this.properties[this.propertiesIndex].PropertyType;
                }
                if (this.fieldsIndex < this.fields.Length)
                {
                    return this.fields[this.fieldsIndex].FieldType;
                }

                throw NewCursorDepletedException();
            }
        }

        public string CurrentPropertyName
        {
            get 
            {
                if (propertiesIndex == -1)
                {
                    throw new InvalidOperationException();
                }
                if (this.propertiesIndex < this.properties.Length)
                {
                    return this.properties[this.propertiesIndex].Name;
                }
                if (this.fieldsIndex < this.fields.Length)
                {
                    return this.fields[this.fieldsIndex].Name;
                }

                throw NewCursorDepletedException();
            }
        }

        public object ReadCurrentProperty()
        {
            if (propertiesIndex == -1)
            {
                throw new InvalidOperationException();
            }
            if (this.propertiesIndex < this.properties.Length)
            {
                return this.properties[this.propertiesIndex].GetGetMethod().Invoke(this.Object, null);
            }
            if (this.fieldsIndex < this.fields.Length)
            {
                return this.fields[this.fieldsIndex].GetValue(this.Object);
            }

            throw NewCursorDepletedException();
        }
    }
}
