﻿// <copyright file="ObjectReader.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ObjectReader Class</summary>

namespace Core.Reflection.Readers
{
    using System;
    using System.Collections.Generic;
    using System.Reflection;
    using Core.Reflection.TypeExpressions;
    using TE = Core.Reflection.TypeExpressions.TypeExpression;

    public sealed class ObjectReader : IObjectReader
    {
        private TypeExpression typeExpression = TE.Subtype(TE.VariableReference("type"), TE.ConstantReference(typeof(object)));
        
        public TypeExpression TypeExpression
        {
            get
            {
                return this.typeExpression;
            }
        }

        public IEnumerable<IReadableProperty> GetReadableProperties(object source)
        {
            Type t = source.GetType();
            PropertyInfo[] properties = t.GetProperties(BindingFlags.Instance | BindingFlags.Public);
            FieldInfo[] fields = t.GetFields(BindingFlags.Instance | BindingFlags.Public);
            
            foreach (PropertyInfo property in properties)
            {
                if (property.CanRead)
                {
                    yield return new ObjectReadableProperty(source, property);
                }
            }

            foreach (FieldInfo field in fields)
            {
                yield return new ObjectReadableField(source, field);
            }
        }

        public IReadableProperty GetProperty(object source, string propertyName)
        {
            Type t = source.GetType();
            PropertyInfo property = t.GetProperty(propertyName, System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Instance);
            if (property == null)
            {
                if (property.CanRead)
                {
                    return new ObjectReadableProperty(source, property);
                }
                else
                {
                    return null;
                }
            }
            else
            {
                FieldInfo field = t.GetField(propertyName, BindingFlags.Public | BindingFlags.Instance);
                if (field == null)
                {
                    return new ObjectReadableField(source, field);
                }
                else
                {
                    return null;
                }
            }
        }

        public IReadablePropertyCursor GetPropertyCursor(object source)
        {
            return new ObjectReadablePropertyCursor(source);
        }

        public object ReadPropertyValue(object source, string propertyName)
        {
            Type t = source.GetType();
            PropertyInfo property = t.GetProperty(propertyName, System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Instance);
            if (property == null)
            {
                if (property.CanRead)
                {
                    return property.GetGetMethod().Invoke(source, null);
                }
                else
                {
                    throw new InvalidOperationException("Property does not exist");
                }
            }
            else
            {
                FieldInfo field = t.GetField(propertyName, BindingFlags.Public | BindingFlags.Instance);
                if (field == null)
                {
                    return field.GetValue(source);
                }
                else
                {
                    throw new InvalidOperationException("Property does not exist");
                }
            }
        }
    }
}
