﻿// <copyright file="CodeExpressionBuilder.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the CodeExpressionBuilder Class</summary>

namespace Core.CodeDom
{
    using System;
    using System.CodeDom;
    using System.Collections.Generic;
    using ExpresssionPair = Core.Collections.ImmutablePair<System.CodeDom.CodeExpression, System.CodeDom.CodeExpression>;

    /// <summary>
    /// A class that contains static methods for building common expressions
    /// </summary>
    public static class CodeExpressionBuilder
    {
        /// <summary>
        /// Builds an IEnumerable of CodeBinaryOperatorExpression objects from pairs of CodeExpression objects.
        /// </summary>
        /// <param name="operator">The operator of the CodeBinaryOperatorExpression object.</param>
        /// <param name="pairsOfExpressions">The array of pairs of Expressions.</param>
        /// <returns>An IEnumerable object containing the CodeBinaryOperatorExpression objects build from the pair of objects supplied.</returns>
        public static IEnumerable<CodeBinaryOperatorExpression> CreateBinaryOperatorExpressionFromPairsOfExpressions(CodeBinaryOperatorType @operator, params ExpresssionPair[] pairsOfExpressions)
        {
            return CreateBinaryOperatorExpressionFromPairsOfExpressions(@operator, pairsOfExpressions);
        }

        /// <summary>
        /// Builds an IEnumerable of CodeBinaryOperatorExpression objects from pairs of CodeExpression objects.
        /// </summary>
        /// <param name="operator">The operator of the CodeBinaryOperatorExpression object.</param>
        /// <param name="pairsOfExpressions">The IEnumerable object of pairs of Expressions.</param>
        /// <returns>An IEnumerable object containing the CodeBinaryOperatorExpression objects build from the pair of objects supplied.</returns>
        public static IEnumerable<CodeBinaryOperatorExpression> CreateBinaryOperatorExpressionFromPairsOfExpressions(CodeBinaryOperatorType @operator, IEnumerable<ExpresssionPair> pairsOfExpressions)
        {
            if (pairsOfExpressions == null)
            {
                throw new ArgumentNullException("pairsOfExpressions");
            }

            foreach (var pair in pairsOfExpressions)
            {
                yield return new CodeBinaryOperatorExpression(pair.FirstItem, @operator, pair.SecondItem);
            }
        }

        /// <summary>
        /// Builds a left associative operator CodeExpression object from many CodeExpression objects.
        /// </summary>
        /// <param name="operator">The operator to used in the composition.</param>
        /// <param name="expressions">The array of CodeExpression objects used to compose the final CodeExpression object.</param>
        /// <returns>The composed CodeExpression object.</returns>
        public static CodeExpression CreateLeftAssociativeMultipleOperatorExpression(CodeBinaryOperatorType @operator, params CodeExpression[] expressions)
        {
            return CreateLeftAssociativeMultipleOperatorExpression(@operator, expressions);
        }

        /// <summary>
        /// Builds a left associative operator CodeExpression object from many CodeExpression objects.
        /// </summary>
        /// <param name="operator">The operator to used in the composition.</param>
        /// <param name="expressions">The IEnumerable object of CodeExpression objects used to compose the final CodeExpression object.</param>
        /// <returns>The composed CodeExpression object.</returns>
        public static CodeExpression CreateLeftAssociativeMultipleOperatorExpression(CodeBinaryOperatorType @operator, IEnumerable<CodeExpression> expressions)
        {
            if (expressions == null)
            {
                throw new ArgumentNullException("expressions");
            }

            IEnumerator<CodeExpression> enumerator = expressions.GetEnumerator();
            CodeExpression result;

            enumerator.Reset();

            if (enumerator.MoveNext())
            {
                result = enumerator.Current;
                while (enumerator.MoveNext())
                {
                    result = new CodeBinaryOperatorExpression(result, @operator, enumerator.Current);
                }
            }
            else
            {
                throw new InvalidOperationException();
            }

            return result;
        }

        /// <summary>
        /// Builds a right associative operator CodeExpression object from many CodeExpression objects.
        /// </summary>
        /// <param name="operator">The operator to used in the composition.</param>
        /// <param name="expressions">The array of CodeExpression objects used to compose the final CodeExpression object.</param>
        /// <returns>The composed CodeExpression object.</returns>
        public static CodeExpression CreateRightAssociativeMultipleOperatorExpression(CodeBinaryOperatorType @operator, params CodeExpression[] expressions)
        {
            return CreateRightAssociativeMultipleOperatorExpression(@operator, expressions);
        }

        /// <summary>
        /// Builds a right associative operator CodeExpression object from many CodeExpression objects.
        /// </summary>
        /// <param name="operator">The operator to used in the composition.</param>
        /// <param name="expressions">The IEnumerable object of CodeExpression objects used to compose the final CodeExpression object.</param>
        /// <returns>The composed CodeExpression object.</returns>
        public static CodeExpression CreateRightAssociativeMultipleOperatorExpression(CodeBinaryOperatorType @operator, IEnumerable<CodeExpression> expressions)
        {
            if (expressions == null)
            {
                throw new ArgumentNullException("expressions");
            }

            IEnumerator<CodeExpression> enumerator = expressions.GetEnumerator();
            CodeExpression result;

            enumerator.Reset();

            if (enumerator.MoveNext())
            {
                result = enumerator.Current;
                while (enumerator.MoveNext())
                {
                    result = new CodeBinaryOperatorExpression(enumerator.Current, @operator, result);
                }
            }
            else
            {
                throw new InvalidOperationException();
            }

            return result;
        }

        /// <summary>
        /// Builds a CodeMethodInvokeExpression object to System.Object.ReferenceEquals.
        /// </summary>
        /// <param name="firstParameter">The first parameter to the method.</param>
        /// <param name="secondParameter">The second parameter to the method.</param>
        /// <returns>The CodeMethodInvokeExpression object.</returns>
        public static CodeMethodInvokeExpression CreateObjectReferenceEqualsInvokeExpression(CodeExpression firstParameter, CodeExpression secondParameter)
        {
            if (firstParameter == null)
            {
                throw new ArgumentNullException("firstParameter");
            }

            if (secondParameter == null)
            {
                throw new ArgumentNullException("secondParameter");
            }

            return
                new CodeMethodInvokeExpression(
                    new CodeTypeReferenceExpression(new CodeTypeReference(typeof(object))),
                    "ReferenceEquals",
                    firstParameter,
                    secondParameter);
        }
    }
}
