﻿// <copyright file="CodeTypeMemberBuilder.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the CodeTypeMemberBuilder Class</summary>

namespace Core.CodeDom
{
    using System;
    using System.Collections.Generic;
    using System.Text;
    using System.CodeDom;

    /// <summary>
    /// Class that provides method for creating common CodeTypeMember objects.
    /// </summary>
    public static class CodeTypeMemberBuilder
    {
        /// <summary>
        /// Creates a public sealed class.
        /// </summary>
        /// <returns>The class.</returns>
        public static CodeTypeDeclaration CreatePublicSealedClass()
        {
            CodeTypeDeclaration result = new CodeTypeDeclaration();
            result.TypeAttributes = System.Reflection.TypeAttributes.Public | System.Reflection.TypeAttributes.Sealed;
            result.IsClass = true;
            return result;
        }

        /// <summary>
        /// Create a parameterless constructor.
        /// </summary>
        /// <returns>The constructor.</returns>
        public static CodeConstructor CreatePublicParameterlessConstructor()
        {
            CodeConstructor result = new CodeConstructor();
            result.Attributes = MemberAttributes.Public;
            return result;
        }

        /// <summary>
        /// Creates a condition statement that checks whether a parameter is null.
        /// </summary>
        /// <param name="parameter">The parameter to check if it is null.</param>
        /// <param name="useIdentityEqualityOperator">Indicates if the identity equality operator (== in C#) should be used. If false then object.ReferenceEquals method is used instead.</param>
        /// <returns>The condition statement.</returns>
        public static CodeConditionStatement CreateArgumentNullCheckStatement(CodeParameterDeclarationExpression parameter, bool useIdentityEqualityOperator)
        {
            CodeConditionStatement condition = new CodeConditionStatement();
            CodeArgumentReferenceExpression argumentReference = new CodeArgumentReferenceExpression(parameter.Name);
            CodePrimitiveExpression nullReference = new CodePrimitiveExpression(null);
            if (useIdentityEqualityOperator)
            {
                condition.Condition = new CodeBinaryOperatorExpression(argumentReference, CodeBinaryOperatorType.IdentityEquality, nullReference);
            }
            else
            {
                condition.Condition = CodeExpressionBuilder.CreateObjectReferenceEqualsInvokeExpression(argumentReference, nullReference);
            }

            condition.TrueStatements.Add(
                new CodeThrowExceptionStatement(
                    new CodeObjectCreateExpression(
                        new CodeTypeReference(typeof(ArgumentNullException)),
                        new CodePrimitiveExpression(parameter.Name))));
            return condition;
        }
    }
}
