﻿// <copyright file="ExceptionClassPattern.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ExceptionClassPattern Class</summary>

namespace Core.CodeDom.Pattern
{
    using System.CodeDom;
    using System.Globalization;
    using System;
    using System.Runtime.Serialization;

    /// <summary>
    /// A pattern for building exception classes.
    /// </summary>
    public sealed class ExceptionClassPattern : IPattern<ExceptionClassPatternInput, ExceptionClassPatternOutput>
    {
        /// <summary>
        /// Initializes a new instance of the ExceptionClassPattern class.
        /// </summary>
        public ExceptionClassPattern()
        {
        }

        /// <summary>
        /// Gets the context of this pattern.
        /// </summary>
        public IPatternContext Context
        {
            get;
            set;
        }

        public ExceptionClassPatternOutput ProduceOutput(ExceptionClassPatternInput input)
        {
            ExceptionClassPatternOutput output = new ExceptionClassPatternOutput();
            CodeTypeDeclaration @class = CodeTypeMemberBuilder.CreatePublicSealedClass();
            @class.Name = GetClassName(input.ExceptionName);
            @class.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(SerializableAttribute))));
            @class.BaseTypes.Add(new CodeTypeReference(typeof(Exception)));
            @class.Members.Add(CodeTypeMemberBuilder.CreatePublicParameterlessConstructor());
            @class.Members.Add(CreateMessageConstructor());
            @class.Members.Add(CreateMessageAndInnerExceptionConstructor());
            @class.Members.Add(CreateSerializationConstructor());

            output.Class = @class;
            return output;
        }

        /// <summary>
        /// Gets the identifier for the class declaration.
        /// </summary>
        /// <param name="exceptionName">The name of the Exception.</param>
        /// <returns>An identifier</returns>
        private static string GetClassName(string exceptionName)
        {
            return string.Format(CultureInfo.InvariantCulture, "{0}Exception", exceptionName);
        }

        /// <summary>
        /// Creates the message constructor.
        /// </summary>
        /// <returns>The message constructor.</returns>
        private static CodeConstructor CreateMessageConstructor()
        {
            CodeConstructor messageConstructor = new CodeConstructor();
            CodeParameterDeclarationExpression mp = new CodeParameterDeclarationExpression();
            mp.Name = "message";
            mp.Type = new CodeTypeReference(typeof(string));
            messageConstructor.Attributes = MemberAttributes.Public;
            messageConstructor.BaseConstructorArgs.Add(new CodeArgumentReferenceExpression(mp.Name));
            messageConstructor.Parameters.Add(mp);
            return messageConstructor;
        }

        /// <summary>
        /// Creates the message and innerException constructor.
        /// </summary>
        /// <returns>The message and innerException constructor.</returns>
        private static CodeConstructor CreateMessageAndInnerExceptionConstructor()
        {
            CodeConstructor messageAndInnerExceptionConstructor = new CodeConstructor();
            CodeParameterDeclarationExpression mp = new CodeParameterDeclarationExpression();
            CodeParameterDeclarationExpression ip = new CodeParameterDeclarationExpression();
            mp.Name = "message";
            mp.Type = new CodeTypeReference(typeof(string));
            ip.Name = "inner";
            ip.Type = new CodeTypeReference(typeof(Exception));
            messageAndInnerExceptionConstructor.Attributes = MemberAttributes.Public;
            messageAndInnerExceptionConstructor.BaseConstructorArgs.Add(new CodeArgumentReferenceExpression(mp.Name));
            messageAndInnerExceptionConstructor.BaseConstructorArgs.Add(new CodeArgumentReferenceExpression(ip.Name));
            messageAndInnerExceptionConstructor.Parameters.Add(mp);
            messageAndInnerExceptionConstructor.Parameters.Add(ip);
            return messageAndInnerExceptionConstructor;
        }

        /// <summary>
        /// Creates the serialization constructor.
        /// </summary>
        /// <returns>The serialization constructor.</returns>
        private static CodeConstructor CreateSerializationConstructor()
        {
            CodeConstructor serializationConstructor = new CodeConstructor();
            CodeParameterDeclarationExpression ip = new CodeParameterDeclarationExpression();
            CodeParameterDeclarationExpression cp = new CodeParameterDeclarationExpression();
            ip.Name = "info";
            ip.Type = new CodeTypeReference(typeof(SerializationInfo));
            cp.Name = "context";
            cp.Type = new CodeTypeReference(typeof(StreamingContext));
            serializationConstructor.Attributes = MemberAttributes.Private;
            serializationConstructor.BaseConstructorArgs.Add(new CodeArgumentReferenceExpression(ip.Name));
            serializationConstructor.BaseConstructorArgs.Add(new CodeArgumentReferenceExpression(cp.Name));
            serializationConstructor.Parameters.Add(ip);
            serializationConstructor.Parameters.Add(cp);
            return serializationConstructor;
        }
    }
}
