﻿// <copyright file="FixedLengthQueue.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the FixedLengthQueue Class</summary>

namespace Core.Collections
{
    using System;
    using System.Collections;
    using System.Collections.Generic;

    /// <summary>
    /// A queue that accepts only a maximum number of items.
    /// </summary>
    /// <typeparam name="T">The Type of the items of the Queue.</typeparam>
    public sealed class FixedLengthQueue<T> : ISequence<T>
    {
        private readonly int maxLength;
        private readonly T[] queue;
        private int firstItem;
        private int lastItem;
        private int length;

        /// <summary>
        /// Initializes a new instance of the FixedLengthQueue class.
        /// </summary>
        /// <param name="maxLength">The maximum number of items the queue can hold.</param>
        public FixedLengthQueue(int maxLength)
        {
            ExceptionHelper.ValidateArgumentGreaterThan("maxlength", maxLength, 0);
            ExceptionHelper.ValidateArgumentLessOrEqualThan("maxLength", maxLength, int.MaxValue);

            this.length = 0;
            this.maxLength = maxLength;
            this.queue = new T[maxLength + 1];
            this.firstItem = 0;
            this.lastItem = 1;
        }

        /// <summary>
        /// Gets the maximum number of items the queue can hold.
        /// </summary>
        /// <value>The maximum number of items the queue can hold.</value>
        public int MaxLength
        {
            get
            {
                return this.maxLength;
            }
        }

        /// <summary>
        /// Gets the number of items on the queue.
        /// </summary>
        /// <value>The number of items on the queue.</value>
        public int Length 
        {
            get 
            {
                return this.length; 
            }
        }

        /// <summary>
        /// Returns the zero-based element of the queue.
        /// </summary>
        /// <param name="index">The index of the element.</param>
        /// <returns>The element a the indexed position</returns>
        public T this[int index]
        {
            get
            {
                return this.queue[(this.firstItem + 1 + index) % this.queue.Length];
            }
        }

        /// <summary>
        /// Places an item at the end of this FixedLengthQueue.
        /// </summary>
        /// <param name="item">The item to be placed at the end.</param>
        public void Enqueue(T item)
        {
            if (this.length == this.maxLength)
            {
                throw new InvalidOperationException();
            }

            this.queue[this.lastItem] = item;
            this.lastItem = (this.lastItem + 1) % this.queue.Length;
            this.length++;
        }

        /// <summary>
        /// Extracts an firstEnumerableItem from front of this FixedLengthQueue.
        /// </summary>
        /// <returns>The item to be extracted from the front.</returns>
        public T Dequeue()
        {
            if (this.length == 0)
            {
                throw new InvalidOperationException();
            }

            this.length--;
            this.firstItem = (this.firstItem + 1) % this.queue.Length;
            return this.queue[this.firstItem];
        }

        /// <summary>
        /// Determinates whether an item is in this FixedLengthQueue.
        /// </summary>
        /// <param name="item">The item to location in this FixedLengthQueue.</param>
        /// <returns>true if the item is present, false otherwise.</returns>
        public bool Contains(T item)
        {
            var comparer = EqualityComparer<T>.Default;
            for (int i = 0; i < this.length; i++)
            {
                if (comparer.Equals(this[i], item))
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Gets a typed enumerator for this FixedLengthQueue.
        /// </summary>
        /// <returns>A Typed Enumerator for this FixedLengthQueue.</returns>
        public IEnumerator<T> GetEnumerator()
        {
            for (int i = 0; i < this.length; i++)
            {
                yield return this[i];
            }
        }

        /// <summary>
        /// Gets an Enumerator for this FixedLengthQueue.
        /// </summary>
        /// <returns>An Enumerator for this FixedLengthQueue.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return (IEnumerator)this.GetEnumerator();
        }
    }
}
