﻿// <copyright file="FixedLookAheadBuffer.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the FixedLookAheadBuffer Class</summary>

namespace Core.Collections
{
    using System;
    using System.Collections.Generic;

    /// <summary>
    /// A fixed length buffer around an IEnumerable object.
    /// </summary>
    /// <typeparam name="T">The Type of the elements of the buffer.</typeparam>
    public sealed class FixedLookAheadBuffer<T> : IDisposable
    {
        private readonly IEnumerable<T> enumerable;
        
        /// <summary>
        /// The amount of items to "look ahead".
        /// </summary>
        private readonly int lookAhead;
        
        /// <summary>
        /// The that will be used to fill the buffer when no more items are found in the IEnumerable object.
        /// </summary>
        private readonly T fillItem;

        private readonly FixedLengthQueue<T> buffer;
        private IEnumerator<T> enumerator;
        private bool endOfBuffer;
        private bool disposed;

        /// <summary>
        /// Initializes a new instance of the FixedLookAheadBuffer class.
        /// </summary>
        /// <param name="enumerable">An IEnumerable.</param>
        /// <param name="lookAhead">The Number of items of look ahead.</param>
        /// <param name="fillItem">The item to yeild in case there are no more items in the enumerable.</param>
        public FixedLookAheadBuffer(IEnumerable<T> enumerable, int lookAhead, T fillItem)
        {
            if (lookAhead >= int.MaxValue - 1)
            {
                throw new ArgumentNullException("lookAhead");
            }

            this.enumerable = enumerable;
            this.lookAhead = lookAhead;
            this.fillItem = fillItem;
            this.buffer = new FixedLengthQueue<T>(lookAhead + 1);
            this.FillBuffer();
        }

        /// <summary>
        /// Finalizes an instance of the FixedLookAheadBuffer class.
        /// </summary>
        ~FixedLookAheadBuffer()
        {
            this.Dispose(false);
        }

        /// <summary>
        /// Gets the next item.
        /// </summary>
        /// <returns>The next item.</returns>
        public T GetItem()
        {
            return this.GetItem(0);
        }

        /// <summary>
        /// Advances one item.
        /// </summary>
        public void Move()
        {
            this.Move(1);
        }

        /// <summary>
        /// Advances a given number of items.
        /// </summary>
        /// <param name="count">The number of items to move forward.</param>
        public void Move(int count)
        {
            for (int i = 0; i < count; i++)
            {
                this.buffer.Dequeue();
                this.buffer.Enqueue(this.GetNextItem());
            }
        }

        /// <summary>
        /// Determines whether the end of the enumerable has been reached.
        /// </summary>
        /// <returns>true if there ar no more items in the IEnumerable, false otherwise.</returns>
        public bool EndOfBuffer()
        {
            return this.endOfBuffer;
        }

        /// <summary>
        /// Gets the item at the given index.
        /// </summary>
        /// <param name="index">The index of the item.</param>
        /// <returns>The item at the given location.</returns>
        public T GetItem(int index) 
        {
            if (index > this.lookAhead)
            {
                throw new InvalidOperationException();
            }

            return this.buffer[index];
        }

        /// <summary>
        /// Resets the Buffer.
        /// </summary>
        public void Reset()
        {
            this.enumerator.Dispose();
            this.FillBuffer();
        }

        /// <summary>
        /// Releases all the resource held by this object.
        /// </summary>
        public void Dispose()
        {
            GC.SuppressFinalize(this);
            this.Dispose(true);
        }

        private T GetNextItem()
        {
            this.endOfBuffer = !this.enumerator.MoveNext();
            if (!this.endOfBuffer)
            {
                return this.enumerator.Current;
            }
            else
            {
                return this.fillItem;
            }
        }

        private void FillBuffer()
        {
            this.endOfBuffer = false;
            this.enumerator = this.enumerable.GetEnumerator();
            while (this.buffer.Length < this.buffer.MaxLength)
            {
                this.buffer.Enqueue(this.GetNextItem());
            }
        }

        private void Dispose(bool disposing)
        {
            if (this.disposed)
            {
                throw new ObjectDisposedException(typeof(FixedLookAheadBuffer<T>).ToString());
            }
            else
            {
                if (disposing)
                {
                    // No unmanged resources to free
                }

                this.enumerator.Dispose();
                this.disposed = true;
            }
        }
    }
}
