﻿// <copyright file="IEnumerableHelper.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the IEnumerableHelper Class</summary>

namespace Core.Collections
{
    using System;
    using System.Collections.Generic;

    /// <summary>
    /// A class that contains static methods for IEnumerable objects.
    /// </summary>
    public static partial class IEnumerableHelper
    {
        /// <summary>
        /// Determines if an item is present in an IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable object.</typeparam>
        /// <param name="enumerable">The IEnumerable object.</param>
        /// <param name="item">The item to look for.</param>
        /// <returns>true if the item is present in the IEnumerable object, false otherwise.</returns>
        public static bool Contains<T>(IEnumerable<T> enumerable, T item)
        {
            if (enumerable == null)
            {
                throw new ArgumentNullException("enumerable");
            }

            return NonValidating.Contains(enumerable, item);
        }

        /// <summary>
        /// Counts the items of an IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable object.</typeparam>
        /// <param name="enumerable">The IEnumerable object.</param>
        /// <returns>The number of items the IEnumerable object has.</returns>
        public static int Count<T>(IEnumerable<T> enumerable)
        {
            if (enumerable == null)
            {
                throw new ArgumentNullException("enumerable");
            }

            return NonValidating.Count(enumerable);
        }

        /// <summary>
        /// Enumerates an IEnumerable object as a set (no repeated elements).
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable object.</typeparam>
        /// <param name="enumerable">The IEnumerable object.</param>
        /// <returns>An IEnumerable object with no repeated elements.</returns>
        public static List<T> EnumerateSetOnList<T>(IEnumerable<T> enumerable)
        {
            if (enumerable == null)
            {
                throw new ArgumentNullException("enumerable");
            }

            return NonValidating.EnumerateSetOnList(enumerable);
        }

        /// <summary>
        /// Determines if an IEnumerable object taken as set is equal to another IEnumerable object taken as set.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>true if the both IEnumerable objects taken as sets are equal, false otherwise.</returns>
        public static bool SetEqual<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.SetEqual(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Determines if an IEnumerable contains the same items in the same order of another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>true if the both IEnumerable objects contains the same items in the same order, false otherwise.</returns>
        public static bool SequenceEqual<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.SequenceEqual(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Determines if an IEnumerable object is a subset of another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>true if the first IEnumerable object is a subset of the second IEnumerable object, false otherwise.</returns>
        public static bool IsSubsetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.IsSubsetOf(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Determines if an IEnumerable object is a superset of another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>true if the first IEnumerable object is a superset of the second IEnumerable object, false otherwise.</returns>
        public static bool IsSupersetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.IsSupersetOf(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Determines if an IEnumerable object is a proper subset of another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>true if the first IEnumerable object is a proper subset of the second IEnumerable object, false otherwise.</returns>
        public static bool IsProperSubsetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.IsProperSubsetOf(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Determines if an IEnumerable object is a proper superset of another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>true if the first IEnumerable object is a proper superset of the second IEnumerable object, false otherwise.</returns>
        public static bool IsProperSupersetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.IsProperSupersetOf(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Determines if the two IEnumerable objects contains items in common.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>true if there is an common item in both IEnumerable objects, false otherwise.</returns>
        public static bool OverlapsWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.OverlapsWith(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Enumerates the difference between an IEnumerable object and another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>The difference between the two IEnumerable objects.</returns>
        public static IEnumerable<T> EnumerateDifferenceWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.EnumerateDifferenceWith(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Enumerates the intersection between an IEnumerable object and another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>The intersection between the two IEnumerable objects.</returns>
        public static IEnumerable<T> EnumerateIntersectionWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.EnumerateIntersectionWith(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// Enumerates the symmetric difference between an IEnumerable object and another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>The symmetric difference between the two IEnumerable objects.</returns>
        public static IEnumerable<T> EnumerateSymmetricDifferenceWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.EnumerateSymmetricDifferenceWith(firstEnumerable, secondEnumerable);
        }
                
        /// <summary>
        /// Enumerates the union between an IEnumerable object and another IEnumerable object.
        /// </summary>
        /// <typeparam name="T">The type of the items in the IEnumerable objects.</typeparam>
        /// <param name="firstEnumerable">The first IEnumerable object.</param>
        /// <param name="secondEnumerable">The second IEnumerable object.</param>
        /// <returns>The union between the two IEnumerable objects.</returns>
        public static IEnumerable<T> EnumerateUnionWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
        {
            if (firstEnumerable == null)
            {
                throw new ArgumentNullException("firstEnumerable");
            }

            if (secondEnumerable == null)
            {
                throw new ArgumentNullException("secondEnumerable");
            }

            return NonValidating.EnumerateUnionWith(firstEnumerable, secondEnumerable);
        }

        /// <summary>
        /// A class that contains static methods for IEnumerable objects that do not validate input.
        /// </summary>
        internal static partial class NonValidating
        {
            /// <summary>
            /// Determines if an item is present in an IEnumerable object.
            /// </summary>
            /// <typeparam name="T">The type of the items in the IEnumerable object.</typeparam>
            /// <param name="enumerable">The IEnumerable object.</param>
            /// <param name="item">The item to look for.</param>
            /// <returns>true if the item is present in the IEnumerable object, false otherwise.</returns>
            internal static bool Contains<T>(IEnumerable<T> enumerable, T item)
            {
                IEqualityComparer<T> comparer = EqualityComparer<T>.Default;
                foreach (T enumerableItem in enumerable)
                {
                    if (comparer.Equals(enumerableItem, item))
                    {
                        return true;
                    }
                }

                return false;
            }

            /// <summary>
            /// Counts the items of an IEnumerable object.
            /// </summary>
            /// <typeparam name="T">The type of the items in the IEnumerable object.</typeparam>
            /// <param name="enumerable">The IEnumerable object.</param>
            /// <returns>The number of items the IEnumerable object has.</returns>
            internal static int Count<T>(IEnumerable<T> enumerable)
            {
                int count = 0;
                IEnumerator<T> enumerator = enumerable.GetEnumerator();
                enumerator.Reset();

                while (enumerator.MoveNext())
                {
                    count++;
                }

                return count;
            }

            internal static List<T> EnumerateSetOnList<T>(IEnumerable<T> enumerable)
            {
                List<T> result = new List<T>();

                foreach (T item in enumerable)
                {
                    if (!result.Contains(item))
                    {
                        result.Add(item);
                    }
                }

                return result;
            }

            internal static bool SetEqual<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                var comparer = EqualityComparer<T>.Default;

                int firstEnumerableLength = 0;
                int secondEnumerableLength = 0;
                foreach (T firstEnumerableItem in firstEnumerable)
                {
                    firstEnumerableLength += 1;
                    foreach (T secondEnumerableItem in secondEnumerable)
                    {
                        if (firstEnumerableLength == 1)
                        {
                            secondEnumerableLength += 1;
                        }

                        if (!comparer.Equals(firstEnumerableItem, secondEnumerableItem))
                        {
                            return false;
                        }
                    }
                }

                if (firstEnumerableLength == secondEnumerableLength)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }

            internal static bool SequenceEqual<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                IEnumerator<T> firstEnumerator = firstEnumerable.GetEnumerator();
                IEnumerator<T> secondEnumerator = secondEnumerable.GetEnumerator();

                IEqualityComparer<T> comparer = EqualityComparer<T>.Default;
                while (firstEnumerator.MoveNext())
                {
                    if (secondEnumerator.MoveNext())
                    {
                        if (!comparer.Equals(firstEnumerator.Current, secondEnumerator.Current))
                        {
                            return false;
                        }
                    }
                    else
                    {
                        return false;
                    }
                }

                if (secondEnumerator.MoveNext())
                {
                    return false;
                }

                return true;
            }

            internal static bool IsSubsetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                foreach (T item in firstEnumerable)
                {
                    if (!Contains(secondEnumerable, item))
                    {
                        return false;
                    }
                }

                return true;
            }

            internal static bool IsSupersetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                return IsSubsetOf(secondEnumerable, firstEnumerable);
            }

            internal static bool IsProperSubsetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                var comparer = EqualityComparer<T>.Default;

                int firstEnumerableLength = 0;
                int secondEnumerableLength = 0;
                foreach (T firstEnumerableItem in firstEnumerable)
                {
                    firstEnumerableLength += 1;
                    foreach (T secondEnumerableItem in secondEnumerable)
                    {
                        if (firstEnumerableLength == 1)
                        {
                            secondEnumerableLength += 1;
                        }

                        if (!comparer.Equals(firstEnumerableItem, secondEnumerableItem))
                        {
                            return false;
                        }
                    }

                    if (firstEnumerableLength >= secondEnumerableLength)
                    {
                        return false;
                    }
                }

                return true;
            }

            internal static bool IsProperSupersetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                return IsProperSubsetOf(secondEnumerable, firstEnumerable);
            }

            internal static bool OverlapsWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                foreach (T item in firstEnumerable)
                {
                    if (Contains(secondEnumerable, item))
                    {
                        return true;
                    }
                }

                return false;
            }

            internal static IEnumerable<T> EnumerateDifferenceWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                foreach (T item in firstEnumerable)
                {
                    if (!Contains(secondEnumerable, item))
                    {
                        yield return item;
                    }
                }
            }

            internal static IEnumerable<T> EnumerateIntersectionWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                foreach (T item in firstEnumerable)
                {
                    if (Contains(secondEnumerable, item))
                    {
                        yield return item;
                    }
                }
            }

            internal static IEnumerable<T> EnumerateSymmetricDifferenceWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                foreach (T item in firstEnumerable)
                {
                    if (!Contains(secondEnumerable, item))
                    {
                        yield return item;
                    }
                }

                foreach (T item in secondEnumerable)
                {
                    if (!Contains(firstEnumerable, item))
                    {
                        yield return item;
                    }
                }
            }

            internal static IEnumerable<T> EnumerateUnionWith<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                foreach (T item in firstEnumerable)
                {
                    if (!Contains(secondEnumerable, item))
                    {
                        yield return item;
                    }
                }

                foreach (T item in secondEnumerable)
                {
                    yield return item;
                }
            }
        }
    }
}
