﻿// <copyright file="ISequenceHelper.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ISequenceHelper Class</summary>

namespace Core.Collections
{
    using System.Collections.Generic;
    using System;

    /// <summary>
    /// A class that contains static methods for ISequence objects.
    /// </summary>
    public static class ISequenceHelper
    {
        /// <summary>
        /// Compares two sequences to see if they are equal.
        /// </summary>
        /// <typeparam name="T">The type of the items in the sequence.</typeparam>
        /// <param name="firstSequence">The first sequence.</param>
        /// <param name="secondSequence">The second sequence.</param>
        /// <returns>true if both sequences contains the same items and in the same order.</returns>
        public static bool SequenceEqual<T>(ISequence<T> firstSequence, ISequence<T> secondSequence)
        {
            if (firstSequence == null)
            {
                throw new ArgumentNullException("firstSequence");
            }

            if (secondSequence == null)
            {
                throw new ArgumentNullException("secondSequence");
            }

            return NonValidating.SequenceEqual(firstSequence, secondSequence);
        }

        /// <summary>
        /// A class that contains static methods for ISequence objects that do not validate input.
        /// </summary>
        internal static class NonValidating
        {
            internal static bool SequenceEqual<T>(ISequence<T> firstSequence, ISequence<T> secondSequence)
            {
                if (firstSequence.Length != secondSequence.Length)
                {
                    return false;
                }

                return IEnumerableHelper.NonValidating.SequenceEqual(firstSequence, secondSequence);
            }

            internal static bool IsSubsetOf<T>(ISequence<T> firstSequence, ISequence<T> secondSequence)
            {
                if (firstSequence.Length > secondSequence.Length)
                {
                    return false;
                }

                return IEnumerableHelper.NonValidating.IsSubsetOf(firstSequence, secondSequence);
            }

            internal static bool IsSupersetOf<T>(ISequence<T> firstSequence, ISequence<T> secondSequence)
            {
                return IsSubsetOf(secondSequence, firstSequence);
            }

            internal static bool SetEqual<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                var comparer = EqualityComparer<T>.Default;

                int firstEnumerableLength = 0;
                int secondEnumerableLength = 0;
                foreach (T firstEnumerableItem in firstEnumerable)
                {
                    firstEnumerableLength += 1;
                    foreach (T secondEnumerableItem in secondEnumerable)
                    {
                        if (firstEnumerableLength == 1)
                        {
                            secondEnumerableLength += 1;
                        }

                        if (!comparer.Equals(firstEnumerableItem, secondEnumerableItem))
                        {
                            return false;
                        }
                    }
                }

                if (firstEnumerableLength == secondEnumerableLength)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }

            internal static bool IsProperSubsetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                var comparer = EqualityComparer<T>.Default;

                int firstEnumerableLength = 0;
                int secondEnumerableLength = 0;
                foreach (T firstEnumerableItem in firstEnumerable)
                {
                    firstEnumerableLength += 1;
                    foreach (T secondEnumerableItem in secondEnumerable)
                    {
                        if (firstEnumerableLength == 1)
                        {
                            secondEnumerableLength += 1;
                        }

                        if (!comparer.Equals(firstEnumerableItem, secondEnumerableItem))
                        {
                            return false;
                        }
                    }

                    if (firstEnumerableLength >= secondEnumerableLength)
                    {
                        return false;
                    }
                }

                return true;
            }

            internal static bool IsProperSupersetOf<T>(IEnumerable<T> firstEnumerable, IEnumerable<T> secondEnumerable)
            {
                return IsProperSubsetOf(secondEnumerable, firstEnumerable);
            }

            internal static bool OverlapsWith<T>(ISequence<T> firstSequence, ISequence<T> secondSequence)
            {
                foreach (T item in firstSequence)
                {
                    if (secondSequence.Contains(item))
                    {
                        return true;
                    }
                }

                return false;
            }
        }
    }
}
