﻿// <copyright file="ImmutableDictionary.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutableDictionary Class</summary>

namespace Core.Collections
{
    using System;
    using System.Collections.Generic;

    /// <summary>
    /// An immutable dictionary.
    /// </summary>
    /// <typeparam name="TKey">The type of the key item.</typeparam>
    /// <typeparam name="TValue">The type of the value.</typeparam>
    public sealed class ImmutableDictionary<TKey, TValue> : IEquatable<ImmutableDictionary<TKey, TValue>>
    {
        /// <summary>
        /// Holds the items of this ImmutableDictionary object.
        /// </summary>
        private readonly Dictionary<TKey, TValue> dictionary = new Dictionary<TKey, TValue>();

        /// <summary>
        /// Holds the keys of this ImmutableDictionary object.
        /// </summary>
        private readonly ImmutableArraySet<TKey> keys;

        /// <summary>
        /// Initializes a new instance of the ImmutableDictionary class.
        /// </summary>
        /// <param name="items">The items for the dictionary.</param>
        public ImmutableDictionary(IDictionary<TKey, TValue> items)
        {
            List<TKey> keysList = new List<TKey>();

            foreach (var key in items.Keys)
            {
                keysList.Add(key);
                this.dictionary.Add(key, items[key]);
            }

            this.keys = new ImmutableArraySet<TKey>(keysList);
        }

        /// <summary>
        /// Initializes a new instance of the ImmutableDictionary class.
        /// </summary>
        /// <param name="items">The items for the dictionary.</param>
        public ImmutableDictionary(ImmutableSet<ImmutablePair<TKey, TValue>> items)
        {
            List<TKey> keysList = new List<TKey>();

            foreach (var element in items)
            {
                keysList.Add(element.FirstItem);
                this.dictionary.Add(element.FirstItem, element.SecondItem);
            }

            this.keys = new ImmutableArraySet<TKey>(keysList);
        }

        /// <summary>
        /// Gets the keys in the dictionary.
        /// </summary>
        /// <value>The key in the dictionary.</value>
        public ImmutableArraySet<TKey> Keys
        {
            get
            {
                return this.keys;
            }
        }

        /// <summary>
        /// Gets the value indexed by the key.
        /// </summary>
        /// <param name="key">The key of the value.</param>
        /// <returns>The value indexed by the key.</returns>
        public TValue this[TKey key]
        {
            get
            {
                return this.GetValue(key);
            }
        }

        /// <summary>
        /// Gets the value indexed by the key.
        /// </summary>
        /// <param name="key">The key of the value.</param>
        /// <returns>The value indexed by the key.</returns>
        public TValue GetValue(TKey key)
        {
            return this.dictionary[key];
        }

        /// <summary>
        /// Compares this ImmutableDictionary against other ImmutableDictionary for equality.
        /// </summary>
        /// <param name="other">The other ImmutableDictionary.</param>
        /// <returns>true if both are equal, false otherwise.</returns>
        public bool Equals(ImmutableDictionary<TKey, TValue> other)
        {
            if (object.ReferenceEquals(other, null))
            {
                return false;
            }
            else
            {
                if (this.keys.Equals(other.keys))
                {
                    foreach (var key in this.keys)
                    {
                        if (!EqualityComparer<TValue>.Default.Equals(this.dictionary[key], other.dictionary[key]))
                        {
                            return false;
                        }
                    }

                    return true;
                }
                else
                {
                    return false;
                }
            }
        }

        /// <summary>
        /// Compares this ImmutableDictionary. against an object for equality.
        /// </summary>
        /// <param name="obj">The object to compare.</param>
        /// <returns>true if this ImmutableDictionary. is equal to the object, false otherwise.</returns>
        public override bool Equals(object obj)
        {
            return this.Equals(obj as ImmutableDictionary<TKey, TValue>);
        }

        /// <summary>
        /// Computes a hash code of this ImmutableDictionary.
        /// </summary>
        /// <returns>The hash code value.</returns>
        public override int GetHashCode()
        {
            return this.keys.GetHashCode();
        }
    }
}
