﻿// <copyright file="ImmutablePair.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutablePair Class</summary>

namespace Core.Collections
{
    using System;
    using System.Collections.Generic;

    /// <summary>
    /// An immutable pair of items.
    /// </summary>
    /// <typeparam name="TFirst">The type of the first item in the pair.</typeparam>
    /// <typeparam name="TSecond">The type of the second item in the pair.</typeparam>
    public sealed class ImmutablePair<TFirst, TSecond> : IEquatable<ImmutablePair<TFirst, TSecond>>
    {
        /// <summary>
        /// Initializes a new instance of the ImmutablePair class.
        /// </summary>
        /// <param name="firstItem">The first item.</param>
        /// <param name="secondItem">The second item.</param>
        public ImmutablePair(TFirst firstItem, TSecond secondItem)
        {
            this.FirstItem = firstItem;
            this.SecondItem = secondItem;
        }

        /// <summary>
        /// Gets the first item.
        /// </summary>
        /// <value>The first item.</value>
        public TFirst FirstItem
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the second item.
        /// </summary>
        /// <value>The second item.</value>
        public TSecond SecondItem
        {
            get;
            private set;
        }

        /// <summary>
        /// Compares two ImmutablePair objects for equality.
        /// </summary>
        /// <param name="firstOperand">The first ImmutablePair object.</param>
        /// <param name="secondOperand">The second ImmutablePair object.</param>
        /// <returns>true if both ImmutablePair objects are equal, false otherwise.</returns>
        public static bool operator ==(ImmutablePair<TFirst, TSecond> firstOperand, ImmutablePair<TFirst, TSecond> secondOperand)
        {
            if (object.ReferenceEquals(firstOperand, null))
            {
                return object.ReferenceEquals(secondOperand, null);
            }
            else
            {
                return firstOperand.Equals(secondOperand);
            }
        }

        /// <summary>
        /// Compares two ImmutablePair objects for inequality.
        /// </summary>
        /// <param name="firstOperand">The first ImmutablePair object.</param>
        /// <param name="secondOperand">The second ImmutablePair object.</param>
        /// <returns>true if both ImmutablePair objects are unequal, false otherwise.</returns>
        public static bool operator !=(ImmutablePair<TFirst, TSecond> firstOperand, ImmutablePair<TFirst, TSecond> secondOperand)
        {
            return !(firstOperand == secondOperand);
        }

        /// <summary>
        /// Compares this ImmutablePair object with another ImmutablePair object for equality.
        /// </summary>
        /// <param name="other">The other ImmutablePair object.</param>
        /// <returns>true if both ImmutablePair objects are equal, false otherwise.</returns>
        public bool Equals(ImmutablePair<TFirst, TSecond> other)
        {
            if (object.ReferenceEquals(other, null))
            {
                return false;
            }
            else
            {
                return EqualityComparer<TFirst>.Default.Equals(this.FirstItem, other.FirstItem) && EqualityComparer<TSecond>.Default.Equals(this.SecondItem, other.SecondItem);
            }
        }

        /// <summary>
        /// Compares this ImmutablePair object with another object for equality.
        /// </summary>
        /// <param name="obj">The other object.</param>
        /// <returns>true if this ImmutablePair object is equal to the other object, false otherwise.</returns>
        public override bool Equals(object obj)
        {
            return this.Equals(obj as ImmutablePair<TFirst, TSecond>);
        }

        /// <summary>
        /// Computes a hash code of this ImmutablePair object.
        /// </summary>
        /// <returns>The hash code value.</returns>
        public override int GetHashCode()
        {
            if (object.ReferenceEquals(this.FirstItem, null))
            {
                if (object.ReferenceEquals(this.SecondItem, null))
                {
                    return 0;
                }
                else
                {
                    return this.SecondItem.GetHashCode();
                }
            }
            else
            {
                if (object.ReferenceEquals(this.SecondItem, null))
                {
                    return this.FirstItem.GetHashCode();
                }
                else
                {
                    return unchecked(this.FirstItem.GetHashCode() + this.SecondItem.GetHashCode());
                }
            }
        }
    }
}
