﻿// <copyright file="ImmutableArraySet.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutableArraySet Class</summary>

namespace Core.Collections
{
    using System.Collections.Generic;

    /// <summary>
    /// An ImmutableSet implemented as an array.
    /// </summary>
    /// <typeparam name="T">The type of the items in the set.</typeparam>
    public sealed class ImmutableArraySet<T> : ImmutableSet<T>
    {
        /// <summary>
        /// The array that holds the items of the this ImmutableArraySet object.
        /// </summary>
        private readonly T[] items;

        /// <summary>
        /// Initializes a new instance of the ImmutableArraySet class.
        /// </summary>
        /// <param name="items">The IEnumerable object containing the items for the set</param>
        public ImmutableArraySet(IEnumerable<T> items)
        {
            List<T> itemsForSet = new List<T>();
            int hc = 0;
            foreach (T item in items)
            {
                if (!itemsForSet.Contains(item))
                {
                    itemsForSet.Add(item);
                    unchecked
                    {
                        hc += item.GetHashCode();
                    }
                }
            }

            this.items = itemsForSet.ToArray();
            this.SetHashCode(hc);
        }

        /// <summary>
        /// Gets the number of items in this ImmutableArraySet.
        /// </summary>
        /// <value>The number of items in this ImmutableArraySet.</value>
        public override int Length
        {
            get
            {
                return this.items.Length;
            }
        }

        /// <summary>
        /// Gets the item at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index of the item to get.</param>
        /// <returns>The item at the specified index.</returns>
        public T this[int index]
        {
            get
            {
                return this.GetItemAt(index);
            }
        }

        /// <summary>
        /// Determines whether an item is present in this ImmutableArraySet.
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <returns>true if the this ImmutableArraySet contains the item, false otherwise.</returns>
        public override bool Contains(T item)
        {
            EqualityComparer<T> comparer = EqualityComparer<T>.Default;

            for (int i = 0; i < this.items.Length; i++)
            {
                if (comparer.Equals(item, this.items[i]))
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Gets a typed enumerator for this ImmutableArraySet object.
        /// </summary>
        /// <returns>A typed enumerator for this ImmutableArraySet object.</returns>
        public override IEnumerator<T> GetEnumerator()
        {
            for (int i = 0; i < this.items.Length; i++)
            {
                yield return this.items[i];
            }
        }

        /// <summary>
        /// Gets the item at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index of the item to get.</param>
        /// <returns>The item at the specified index.</returns>
        public T GetItemAt(int index)
        {
            return this.items[index];
        }
    }
}
