﻿// <copyright file="ImmutableCartesianProductSet.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutableCartesianProductSetElementBuilder Delegate and the ImmutableCartesianProductSet Class</summary>

namespace Core.Collections
{
    using System;
    using System.Collections.Generic;

    /// <summary>
    /// Builds the items of an ImmutableCartesianProductSet object.
    /// </summary>
    /// <typeparam name="TCartesianProductSetType">The type of the items in the set.</typeparam>
    /// <typeparam name="TFirstSetType">The type of the items in the first set.</typeparam>
    /// <typeparam name="TSecondSetType">The type of the items in the second set.</typeparam>
    /// <param name="firstElement">An item from the first set.</param>
    /// <param name="secondElement">An item from the second set.</param>
    /// <returns>An item for the cartesian product set</returns>
    public delegate TCartesianProductSetType ImmutableCartesianProductSetElementBuilder<TCartesianProductSetType, TFirstSetType, TSecondSetType>(TFirstSetType firstElement, TSecondSetType secondElement);

    /// <summary>
    /// An immutable cartesian product of two ImmutableSet objects.
    /// </summary>
    /// <typeparam name="TCartesianProductSetType">The type of the items in the set.</typeparam>
    /// <typeparam name="TFirstSetType">The type of the items in the first set.</typeparam>
    /// <typeparam name="TSecondSetType">The type of the items in the second set.</typeparam>
    public sealed class ImmutableCartesianProductSet<TCartesianProductSetType, TFirstSetType, TSecondSetType> : ImmutableSet<TCartesianProductSetType>
    {       
        /// <summary>
        /// Initializes a new instance of the ImmutableCartesianProductSet class.
        /// </summary>
        /// <param name="firstSet">The first ImmutableSet object.</param>
        /// <param name="secondSet">The second ImmutableSet object.</param>
        /// <param name="elementBuilder">The function object used to build the items of this ImmutableCartesianProductSet.</param>
        public ImmutableCartesianProductSet(ImmutableSet<TFirstSetType> firstSet, ImmutableSet<TSecondSetType> secondSet, ImmutableCartesianProductSetElementBuilder<TCartesianProductSetType, TFirstSetType, TSecondSetType> elementBuilder)
        {
            if (firstSet == null)
            {
                throw new ArgumentNullException("firstSet");
            }

            if (secondSet == null)
            {
                throw new ArgumentNullException("secondSet");
            }

            if (elementBuilder == null)
            {
                throw new ArgumentNullException("elementBuilder");
            }

            this.FirstSet = firstSet;
            this.SecondSet = secondSet;
            this.ElementBuilder = elementBuilder;
        }

        /// <summary>
        /// Gets the first ImmutableSet object.
        /// </summary>
        /// <value>The first ImmutableSet object.</value>
        public ImmutableSet<TFirstSetType> FirstSet 
        { 
            get; 
            private set; 
        }

        /// <summary>
        /// Gets the second ImmutableSet object.
        /// </summary>
        /// <value>The second ImmutableSet object.</value>
        public ImmutableSet<TSecondSetType> SecondSet 
        {
            get;
            private set; 
        }
        
        /// <summary>
        /// Gets the function object used to build the items of this ImmutableCartesianProductSet
        /// </summary>
        /// <value></value>
        public ImmutableCartesianProductSetElementBuilder<TCartesianProductSetType, TFirstSetType, TSecondSetType> ElementBuilder 
        {
            get;
            private set; 
        }

        /// <summary>
        /// Gets a typed enumerator for this ImmutableCartesianProductSet object.
        /// </summary>
        /// <returns>A typed enumerator for this ImmutableCartesianProductSet object.</returns>
        public override IEnumerator<TCartesianProductSetType> GetEnumerator()
        {
            foreach (TFirstSetType firstSetElement in this.FirstSet)
            {
                foreach (TSecondSetType secondSetElement in this.SecondSet)
                {
                    yield return this.ElementBuilder(firstSetElement, secondSetElement);
                }
            }
        }
    }
}
