﻿// <copyright file="ImmutableIntersectionSet.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutableIntersectionSet Class</summary>

namespace Core.Collections
{
    using System.Collections.Generic;
    using System;

    /// <summary>
    /// An immutable intersection of two ImmutableSet objects.
    /// </summary>
    /// <typeparam name="T">The type of the items in the set.</typeparam>
    public sealed class ImmutableIntersectionSet<T> : ImmutableSet<T>
    {
        /// <summary>
        /// The IEnumerable object used to hold the enumeration of the items of this set.
        /// </summary>
        private IEnumerable<T> set;

        /// <summary>
        /// Initializes a new instance of the ImmutableIntersectionSet class.
        /// </summary>
        /// <param name="firstSet">The first ImmutableSet object.</param>
        /// <param name="secondSet">The second ImmutableSet object.</param>
        public ImmutableIntersectionSet(ImmutableSet<T> firstSet, ImmutableSet<T> secondSet)
        {
            if (firstSet == null)
            {
                throw new ArgumentNullException("firstSet");
            }

            if (secondSet == null)
            {
                throw new ArgumentNullException("secondSet");
            }
            
            this.FirstSet = firstSet;
            this.SecondSet = secondSet;
        }

        /// <summary>
        /// Gets the first ImmutableSet object.
        /// </summary>
        /// <value>The first ImmutableSet object.</value>
        public ImmutableSet<T> FirstSet
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the second ImmutableSet object.
        /// </summary>
        /// <value>The second ImmutableSet object.</value>
        public ImmutableSet<T> SecondSet
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets a typed enumerator for this ImmutableIntersectionSet object.
        /// </summary>
        /// <returns>A typed enumerator for this ImmutableIntersectionSet object.</returns>
        public override IEnumerator<T> GetEnumerator()
        {
            if (this.set == null)
            {
                this.set = IEnumerableHelper.NonValidating.EnumerateIntersectionWith(this.FirstSet, this.SecondSet);
            }

            return this.set.GetEnumerator();
        }
    }
}
