﻿// <copyright file="ImmutableSet.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutableSet Class</summary>

namespace Core.Collections
{
    using System;
    using System.Collections;
    using System.Collections.Generic;

    /// <summary>
    /// An immutable set of items.
    /// </summary>
    /// <typeparam name="T">The type of the items in the set.</typeparam>
    public abstract class ImmutableSet<T> : ISequence<T>, IEquatable<ImmutableSet<T>>
    {
        /// <summary>
        /// The empty set object.
        /// </summary>
        private static readonly ImmutableSet<T> emptySet = new ImmutableArraySet<T>(new T[0]);
        
        /// <summary>
        /// Contains the length of this set, -1 means the set has no been measured
        /// </summary>
        private int length = -1;
        
        /// <summary>
        /// Contains the hash code of this set, null means the hash code has not been computed
        /// </summary>
        private int? hashCode;

        /// <summary>
        /// Gets the empty set.
        /// </summary>
        /// <value>The empty set.</value>
        public static ImmutableSet<T> EmptySet
        {
            get
            {
                return emptySet;
            }
        }

        /// <summary>
        /// Gets the number of items in the set.
        /// </summary>
        /// <value>The number of items in the set.</value>
        public virtual int Length
        {
            get
            {
                if (this.length == -1)
                {
                    this.length = IEnumerableHelper.NonValidating.Count(this);
                }

                return this.length;
            }
        }

        /// <summary>
        /// Returns the difference of this set and another set.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>The difference of the two sets.</returns>
        public virtual ImmutableSet<T> DifferenceWith(ImmutableSet<T> other)
        {
            return new ImmutableDifferenceSet<T>(this, other);
        }

        /// <summary>
        /// Returns the intersection of this set and another set.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>The intersection of the two sets.</returns>
        public virtual ImmutableSet<T> IntersectionWith(ImmutableSet<T> other)
        {
            return new ImmutableIntersectionSet<T>(this, other);
        }

        /// <summary>
        /// Determines whether this set is a proper subset of another set.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>true if this set is a proper subset of the other set, false otherwise.</returns>
        public virtual bool IsProperSubsetOf(ImmutableSet<T> other)
        {
            if (other == null)
            {
                throw new ArgumentNullException("other");
            }

            return ISequenceHelper.NonValidating.IsProperSubsetOf(this, other);
        }

        /// <summary>
        /// Determines whether this set is a proper superset of another set.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>true if this set is a proper superset of the other set, false otherwise.</returns>
        public virtual bool IsProperSupersetOf(ImmutableSet<T> other)
        {
            if (other == null)
            {
                throw new ArgumentNullException("other");
            }

            return ISequenceHelper.NonValidating.IsProperSupersetOf(this, other);
        }

        /// <summary>
        /// Determines whether this set is a subset of another set.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>true if this set is a subset of the other set, false otherwise.</returns>
        public virtual bool IsSubsetOf(ImmutableSet<T> other)
        {
            if (other == null)
            {
                throw new ArgumentNullException("other");
            }

            return ISequenceHelper.NonValidating.IsSubsetOf(this, other);
        }

        /// <summary>
        /// Determines whether this set is a superset of another set.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>true if this set is a superset of the other set, false otherwise.</returns>
        public virtual bool IsSupersetOf(ImmutableSet<T> other)
        {
            if (other == null)
            {
                throw new ArgumentNullException("other");
            }
                
            return ISequenceHelper.NonValidating.IsSupersetOf(this, other);
        }

        /// <summary>
        /// Determines if this set and other set have elements in common.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>true if equal, false otherwise.</returns>
        public virtual bool OverlapsWith(ImmutableSet<T> other)
        {
            if (other == null)
            {
                throw new ArgumentNullException("other");
            }
                
            return ISequenceHelper.NonValidating.OverlapsWith(this, other);
        }

        /// <summary>
        /// Returns the symmetric difference of this set and another set.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>The intersection of the two sets.</returns>
        public virtual ImmutableSet<T> SymmetricDifferenceWith(ImmutableSet<T> other)
        {
            return new ImmutableSymmetricDifferenceSet<T>(this, other);
        }

        /// <summary>
        /// Returns the union of this set and another set.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>The intersection of the two sets.</returns>
        public virtual ImmutableSet<T> UnionWith(ImmutableSet<T> other)
        {
            return new ImmutableUnionSet<T>(this, other);
        }

        /// <summary>
        /// Determines whether this set contains an item.
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <returns>true if the item belongs to the set, false otherwise.</returns>
        public virtual bool Contains(T item)
        {
            return IEnumerableHelper.Contains(this, item);
        }

        /// <summary>
        /// Gets a typed enumerator for this ImmutableSet object.
        /// </summary>
        /// <returns>A Typed Enumerator for this ImmutableSet object.</returns>
        public abstract IEnumerator<T> GetEnumerator();

        /// <summary>
        /// Gets an enumerator for this ImmutableSet.
        /// </summary>
        /// <returns>An Enumerator for this ImmutableSet.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            foreach (object item in this)
            {
                yield return item;
            }
        }

        /// <summary>
        /// Compares this set with another set for equality.
        /// </summary>
        /// <param name="other">The other set.</param>
        /// <returns>true if both sets are equal, false otherwise.</returns>
        public bool Equals(ImmutableSet<T> other)
        {
            if (object.ReferenceEquals(other, null))
            {
                return false;
            }

            return ISequenceHelper.NonValidating.SetEqual(this, other);
        }

        /// <summary>
        /// Compares this set against an object for equality.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <returns>true if the set is equal to the object, false otherwise.</returns>
        public override bool Equals(object obj)
        {
            return this.Equals(obj as ImmutableSet<T>);
        }

        /// <summary>
        /// Gets the hash code of this set.
        /// </summary>
        /// <returns>The code of this set.</returns>
        public override int GetHashCode()
        {
            if (!this.hashCode.HasValue)
            {
                this.ComputeHashCode();
            }

            return this.hashCode.Value;
        }

        /// <summary>
        /// Computes the hash code of this set.
        /// </summary>
        protected virtual void ComputeHashCode()
        {
            int hc = 0;

            foreach (T item in this)
            {
                hc += unchecked(item.GetHashCode());
            }

            this.SetHashCode(hc);
        }

        /// <summary>
        /// Sets the hash code of this set.
        /// </summary>
        /// <param name="hc">The hash code.</param>
        protected void SetHashCode(int hc)
        {
            this.hashCode = hc;
        }
    }
}
