﻿// <copyright file="ImmutableArrayString.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutableArrayString Class</summary>

namespace Core.Collections
{
    using System.Collections.Generic;
    using System;

    /// <summary>
    /// An ImmutableString implemented as an array.
    /// </summary>
    /// <typeparam name="T">The type of the items in the string.</typeparam>
    public class ImmutableArrayString<T> : ImmutableString<T>
    {
        /// <summary>
        /// Holds the items of this ImmutableArrayString object.
        /// </summary>
        private readonly T[] str;

        /// <summary>
        /// Initializes a new instance of the ImmutableArrayString class.
        /// </summary>
        /// <param name="array">The Array object containing the items to initialize the string.</param>
        public ImmutableArrayString(params T[] array)
        {
            if (array == null)
            {
                throw new ArgumentNullException("array");
            }

            this.str = ArrayHelper.NonValidating.NewArrayFrom(array);
        }

        /// <summary>
        /// Initializes a new instance of the ImmutableArrayString class.
        /// </summary>
        /// <param name="list">The IList object containing the items to initialize the string.</param>
        public ImmutableArrayString(IList<T> list)
        {
            if (list == null)
            {
                throw new ArgumentNullException("list");
            }

            this.str = ArrayHelper.NonValidating.NewArrayFrom(list);
        }

        /// <summary>
        /// Initializes a new instance of the ImmutableArrayString class.
        /// </summary>
        /// <param name="enumerable">The IEnumerable object containing the items to initialize the string.</param>
        public ImmutableArrayString(IEnumerable<T> enumerable)
        {
            if (enumerable == null)
            {
                throw new ArgumentNullException("enumerable");
            }
                
            this.str = ArrayHelper.NonValidating.NewArrayFrom(enumerable);
        }

        /// <summary>
        /// Initializes a new instance of the ImmutableArrayString class.
        /// </summary>
        /// <param name="firstString">The first ImmutableString object.</param>
        /// <param name="secondString">The second ImmutableString object.</param>
        public ImmutableArrayString(ImmutableString<T> firstString, ImmutableString<T> secondString)
        {

            if (firstString == null)
            {
                throw new ArgumentNullException("firstString");
            }

            if (secondString == null)
            {
                throw new ArgumentNullException("secondString");
            }

            this.str = Concatenate(firstString, secondString);
        }

        /// <summary>
        /// Gets the number of items in this string.
        /// </summary>
        /// <value>The number of items in this string.</value>
        public sealed override int Length
        {
            get
            {
                return this.str.Length;
            }
        }

        /// <summary>
        /// Gets the item at the given location.
        /// </summary>
        /// <param name="index">The zero-based index of the item to get.</param>
        /// <returns>The item at the given location.</returns>
        public sealed override T GetItemAt(int index)
        {
            return this.str[index];
        }

        /// <summary>
        /// Determines whether an item is present in this ImmutableArrayString object..
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <returns>true if the item is present, false otherwise.</returns>
        public sealed override bool Contains(T item)
        {
            IEqualityComparer<T> comparer = EqualityComparer<T>.Default;
            for (int i = 0; i < this.str.Length; i++)
            {
                if (comparer.Equals(this.str[i], item))
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Gets a typed enumerator for this ImmutableArrayString object.
        /// </summary>
        /// <returns>A typed enumerator for this ImmutableArrayString object.</returns>
        public sealed override IEnumerator<T> GetEnumerator()
        {
            return ((IEnumerable<T>)str).GetEnumerator();
        }

        /// <summary>
        /// Creates an array with the concatenation of two strings.
        /// </summary>
        /// <param name="firstString">The first ImmutableString object.</param>
        /// <param name="secondString">The second ImmutableString object.</param>
        /// <returns>An array with the concatenation of two strings.</returns>
        private static T[] Concatenate(ImmutableString<T> firstString, ImmutableString<T> secondString)
        {
            T[] result = new T[firstString.Length + secondString.Length];

            ArrayHelper.NonValidating.FillFrom(result, firstString);

            for (int i = 0, j = firstString.Length; i < secondString.Length; i++, j++)
            {
                result[j] = secondString[i];
            }

            return result;
        }
    }
}
