﻿// <copyright file="ImmutableConcatenationString.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutableConcatenationString Class</summary>

namespace Core.Collections
{
    using System.Collections.Generic;

    /// <summary>
    /// An immutable string of the concatenation of two strings.
    /// </summary>
    /// <typeparam name="T">The type of the items in the string.</typeparam>
    public sealed class ImmutableConcatenationString<T> : ImmutableString<T>
    {
        /// <summary>
        /// Initializes a new instance of the ImmutableConcatenationString class.
        /// </summary>
        /// <param name="firstString">The first ImmutableString object.</param>
        /// <param name="secondString">The second ImmutableString object.</param>
        public ImmutableConcatenationString(ImmutableString<T> firstString, ImmutableString<T> secondString)
        {
            this.FirstString = firstString;
            this.SecondString = secondString;
        }

        /// <summary>
        /// Gets the first ImmutableString object.
        /// </summary>
        /// <value>The first ImmutableString object.</value>
        public ImmutableString<T> FirstString
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the second ImmutableString object.
        /// </summary>
        /// <value>The second ImmutableString object.</value>
        public ImmutableString<T> SecondString
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the number of items in this string.
        /// </summary>
        /// <value>The number of items in this string.</value>
        public override int Length
        {
            get
            {
                return this.FirstString.Length + this.SecondString.Length;
            }
        }

        /// <summary>
        /// Gets the item at the given location.
        /// </summary>
        /// <param name="index">The zero-based index of the item to get.</param>
        /// <returns>The item at the given location.</returns>
        public override T GetItemAt(int index)
        {
            if (index < this.FirstString.Length)
            {
                return this.FirstString.GetItemAt(index);
            }
            else
            {
                return this.SecondString.GetItemAt(index - this.FirstString.Length);
            }
        }

        /// <summary>
        /// Determines whether an item is present in this ImmutableConcatenationString object.
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <returns>true if the item is present, false otherwise.</returns>
        public override bool Contains(T item)
        {
            if (this.FirstString.Contains(item))
            {
                return true;
            }
            else
            {
                return this.SecondString.Contains(item);
            }
        }

        /// <summary>
        /// Gets a typed enumerator for this ImmutableConcatenationString object.
        /// </summary>
        /// <returns>A typed enumerator for this ImmutableConcatenationString object.</returns>
        public override IEnumerator<T> GetEnumerator()
        {
            foreach (T item in this.FirstString)
            {
                yield return item;
            }

            foreach (T item in this.SecondString)
            {
                yield return item;
            }
        }
    }
}
