﻿// <copyright file="ImmutableString.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ImmutableString Class</summary>

namespace Core.Collections
{
    using System;
    using System.Collections;
    using System.Collections.Generic;

    /// <summary>
    /// An immutable string of items.
    /// </summary>
    /// <typeparam name="T">The type of the items in the string.</typeparam>
    public abstract class ImmutableString<T> : IReadableRandomAccessSequence<T>, IEquatable<ImmutableString<T>>
    {
        /// <summary>
        /// The empty string object.
        /// </summary>
        private static readonly ImmutableArrayString<T> emptyString = new ImmutableArrayString<T>();

        /// <summary>
        /// Contains the hash code of this string, null means the hash code has not been computed
        /// </summary>
        private int? hashCode;

        /// <summary>
        /// Gets an empty string of items.
        /// </summary>
        /// <value>An empty string of items.</value>
        public static ImmutableString<T> EmptyString
        {
            get
            {
                return emptyString;         
            }
        }

        /// <summary>
        /// Gets the number of items in this string.
        /// </summary>
        /// <value>The number of items in this string.</value>
        public virtual int Length
        {
            get
            {
                return IEnumerableHelper.NonValidating.Count(this);
            }
        }

        /// <summary>
        /// Gets the item at the given location.
        /// </summary>
        /// <param name="index">The zero-based index of the item to get.</param>
        /// <returns>The item at the given location.</returns>
        public T this[int index]
        {
            get
            {
                return this.GetItemAt(index);
            }
        }

        /// <summary>
        /// Concatenates two ImmutableString"/>.
        /// </summary>
        /// <param name="other">The other string to be concatenated.</param>
        /// <returns>The concatenation of the two strings.</returns>
        public virtual ImmutableString<T> Concatenate(ImmutableString<T> other)
        {
            return new ImmutableConcatenationString<T>(this, other);
        }

        /// <summary>
        /// Determines whether an item is present in this ImmutableString.
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <returns>true if the item is present, false otherwise.</returns>
        public virtual bool Contains(T item)
        {
            return IEnumerableHelper.NonValidating.Contains(this, item);
        }

        /// <summary>
        /// Gets the item at the given location.
        /// </summary>
        /// <param name="index">The zero-based index of the item to get.</param>
        /// <returns>The item at the given location.</returns>
        public abstract T GetItemAt(int index);

        /// <summary>
        /// Gets a typed enumerator for this ImmutableString object.
        /// </summary>
        /// <returns>A typed Enumerator for this ImmutableString object.</returns>
        public abstract IEnumerator<T> GetEnumerator();

        /// <summary>
        /// Gets an enumerator for this ImmutableString object.
        /// </summary>
        /// <returns>An enumerator for this ImmutableString object.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            foreach (object item in this)
            {
                yield return item;
            }
        }

        /// <summary>
        /// Compares this string with another string for equality.
        /// </summary>
        /// <param name="other">The other string.</param>
        /// <returns>true if both strings are equal, false otherwise.</returns>
        public virtual bool Equals(ImmutableString<T> other)
        {
            if (object.ReferenceEquals(other, null))
            {
                return false;
            }

            return ISequenceHelper.NonValidating.SequenceEqual(this, other);
        }

        /// <summary>
        /// Compares this string against an object for equality.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <returns>true if the string is equal to the object, false otherwise.</returns>
        public override bool Equals(object obj)
        {
            return this.Equals(obj as ImmutableString<T>);
        }

        /// <summary>
        /// Gets the hash code of this string.
        /// </summary>
        /// <returns>The code of this string.</returns>
        public override int GetHashCode()
        {
            if (!this.hashCode.HasValue)
            {
                int hc = 0;
                for (int i = 0; i < this.Length; i++)
                {
                    hc = unchecked(hc + this[i].GetHashCode());
                }

                this.hashCode = hc;
            }

            return this.hashCode.Value;
        }
    }
}
