﻿// <copyright file="ReflectionFunctions.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the ReflectionFunctions Class</summary>

namespace Core.Reflection
{
    using System;
    using System.Collections.Generic;
    using System.Reflection;

    public static class ReflectionFunctions
    {
        /// <summary>
        /// Fills a IDictionary&lt;string, object&gt; instance with the values of an object's public instance properties.
        /// </summary>
        /// <param name="dictionary">The dictionary to fill.</param>
        /// <param name="obj">The object where to look for properties.</param>
        public static void FillDictionaryWithPublicInstanceProperties(IDictionary<string, object> dictionary, object @object)
        {
            Type type = @object.GetType();
            PropertyInfo[] properties = type.GetProperties(BindingFlags.Public | BindingFlags.Instance);

            foreach (PropertyInfo property in properties)
            {
                if (property.CanRead)
                {
                    MethodInfo propertyGetMethod = property.GetGetMethod();
                    if (propertyGetMethod.GetParameters().Length == 0)
                    {
                        object value = propertyGetMethod.Invoke(@object, null);
                        if (dictionary.ContainsKey(property.Name))
                        {
                            dictionary[property.Name] = value;
                        }
                        else
                        {
                            dictionary.Add(property.Name, value);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Fills a IDictionary&lt;string, object&gt; instance with the values of an object's public instance fields .
        /// </summary>
        /// <param name="dictionary">The dictionary to fill.</param>
        /// <param name="obj">The object where to look for properties.</param>
        public static void FillDictionaryWithPublicInstanceFields(IDictionary<string, object> dictionary, object @object)
        {
            Type type = @object.GetType();
            FieldInfo[] fields = type.GetFields(BindingFlags.Public | BindingFlags.Instance);

            foreach (FieldInfo field in fields)
            {
                object value = field.GetValue(@object);
                if (dictionary.ContainsKey(field.Name))
                {
                    dictionary[field.Name] = value;
                }
                else
                {
                    dictionary.Add(field.Name, value);
                }
            }
        }

        public static void FillDictionaryWithObject(IDictionary<string, object> dictionary, object @object)
        {
            FillDictionaryWithPublicInstanceProperties(dictionary, @object);
            FillDictionaryWithPublicInstanceFields(dictionary, @object);
        }

        public static void FillPublicInstancePropertiesWithDictionary(object @object, IDictionary<string, object> dictionary)
        {
            Type type = @object.GetType();

            foreach (var propertyName in dictionary.Keys)
            {
                if (propertyName != null)
                {
                    object value = dictionary[propertyName];
                    PropertyInfo property = type.GetProperty(propertyName, BindingFlags.Public | BindingFlags.Instance);
                    if (property.CanWrite)
                    {
                        MethodInfo method = property.GetSetMethod();
                        if (method.GetParameters().Length == 1)
                        {
                            method.Invoke(@object, new object[] { value });
                        }
                    }
                }
            }
        }

        public static void FillPublicInstanceFieldsWithDictionary(object @object, IDictionary<string, object> dictionary)
        {
            Type type = @object.GetType();

            foreach (var fieldName in dictionary.Keys)
            {
                if (fieldName != null)
                {
                    object value = dictionary[fieldName];
                    FieldInfo field = type.GetField(fieldName, BindingFlags.Public | BindingFlags.Instance);
                    field.SetValue(@object, value);
                }
            }
        }

        public static void FillObjectWithDictionary(object @object, IDictionary<string, object> dictionary)
        {
            FillPublicInstancePropertiesWithDictionary(@object, dictionary);
            FillPublicInstanceFieldsWithDictionary(@object, dictionary);
        }
    }
}
