﻿// <copyright file="TypeExpressionEvaluationVisitor.cs" company="Felipe José Angriman">
//   Copyright 2009 Felipe José Angriman 
//
//   Licensed under the Apache License, Version 2.0 (the "License"); 
//   you may not use this file except in compliance with the License. 
//   You may obtain a copy of the License at 
//   
//       http://www.apache.org/licenses/LICENSE-2.0 
// 
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS, 
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and 
//   limitations under the License. 
// </copyright>
// <author>Felipe José Angriman</author>
// <summary>Defines the TypeExpressionEvaluationVisitor Class</summary>

namespace Core.Reflection.TypeExpressions
{
    using System;
    using System.Collections.Generic;

    public sealed class TypeExpressionEvaluationVisitor : ITypeExpressionVisitor
    {
        private IDictionary<string, Type> variables;
        private TypeExpression rootNode;
        private Stack<Type> typeEvaluation;
        private Stack<bool> results;

        public bool Evaluate(TypeExpression rootNode, IDictionary<string, Type> variables)
        {
            this.rootNode = rootNode;
            this.variables = variables;
            this.typeEvaluation = new Stack<Type>();
            this.results = new Stack<bool>();
            this.rootNode.Accept(this);
            return results.Pop();
        }

        public void Visit(TypeAndExpression expression)
        {
            expression.FirstExpression.Accept(this);
            expression.SecondExpression.Accept(this);
            bool secondValue = results.Pop();
            bool firstValue = results.Pop();
            results.Push(firstValue && secondValue);
        }

        public void Visit(TypeOrExpression expression)
        {
            expression.FirstExpression.Accept(this);
            expression.SecondExpression.Accept(this);
            bool secondValue = results.Pop();
            bool firstValue = results.Pop();
            results.Push(firstValue || secondValue);
        }

        public void Visit(TypeNotExpression expression)
        {
            expression.Expression.Accept(this);
            results.Push(!results.Pop());
        }

        public void Visit(TypeConstantReferenceExpression expression)
        {
            typeEvaluation.Push(expression.Constant);
        }

        public void Visit(TypeEqualsExpression expression)
        {
            expression.FirstExpression.Accept(this);
            expression.SecondExpression.Accept(this);
            Type secondType = typeEvaluation.Pop();
            Type firstType = typeEvaluation.Pop();
            results.Push(firstType == secondType);
        }

        public void Visit(TypeSubtypeExpression expression)
        {
            expression.FirstExpression.Accept(this);
            expression.SecondExpression.Accept(this);
            Type secondType = typeEvaluation.Pop();
            Type firstType = typeEvaluation.Pop();
            results.Push(secondType.IsAssignableFrom(firstType));
        }

        public void Visit(TypeVariableReferenceExpression expression)
        {
            typeEvaluation.Push(variables[expression.VariableName]);
        }
        
        public void Visit(TypeIsGenericExpression expression)
        {
            expression.FirstExpression.Accept(this);
            expression.SecondExpression.Accept(this);
            Type secondType = typeEvaluation.Pop();
            Type firstType = typeEvaluation.Pop();
            results.Push(firstType.IsGenericType && firstType.GetGenericTypeDefinition() == secondType);
        }
    }
}
